"""Chat message container base class"""

from __future__ import annotations

import os
from collections.abc import Generator
from contextlib import contextmanager
from dataclasses import dataclass
from datetime import UTC, datetime
from io import StringIO

import orjson as json
import rich.repr

from parllama.chat_message import ParllamaChatMessage
from parllama.messages.par_chat_messages import ParChatMessageDeleted
from parllama.par_event_system import ParEventSystemBase
from parllama.settings_manager import settings


@rich.repr.auto
@dataclass
class ChatMessageContainer(ParEventSystemBase):
    """Chat message container base class"""

    _name: str
    """Name of the chat message container"""
    messages: list[ParllamaChatMessage]
    """Messages in the chat message container"""
    last_updated: datetime
    """Last updated timestamp of the chat message container"""

    _id_to_msg: dict[str, ParllamaChatMessage]
    _changes: set[str]
    _batching: bool
    _loaded: bool

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        *,
        id: str | None = None,  # pylint: disable=redefined-builtin
        name: str | None = None,
        messages: list[ParllamaChatMessage] | list[dict] | None = None,
        last_updated: datetime | None = None,
    ):
        """Initialize the chat prompt"""
        super().__init__(id=id)
        self._batching = True
        self._changes = set()
        self._id_to_msg = {}
        self._name = name or "Messages"
        self.messages = []
        msgs = messages or []
        for m in msgs:
            msg: ParllamaChatMessage
            if isinstance(m, ParllamaChatMessage):
                msg = m
            else:
                msg = ParllamaChatMessage(**m)
            self.messages.append(msg)
            self._id_to_msg[msg.id] = msg
            self.mount(msg)
        self.last_updated = last_updated or datetime.now(UTC)
        self._loaded = messages is not None
        self._batching = False

    def unload(self) -> None:
        """Unload the messages"""
        self._loaded = False
        self.messages = []

    @property
    def is_loaded(self):
        """Check if the messages have been loaded"""
        return self._loaded

    @property
    def batching(self) -> bool:
        """Check if the session is loading"""
        return self._batching

    @batching.setter
    def batching(self, value: bool) -> None:
        """Set the loading state"""
        self._batching = value

    def add_message(self, msg: ParllamaChatMessage, prepend: bool = False) -> None:
        """Add a message"""
        if prepend:
            self.messages.insert(0, msg)
        else:
            self.messages.append(msg)
        self._id_to_msg[msg.id] = msg
        self.last_updated = datetime.now(UTC)
        self.mount(msg)
        self._changes.add("messages")
        self._loaded = True
        # self.log_it(f"Added {msg.role} message to: {self.name}")
        self.save()

    @property
    def name(self) -> str:
        """Get the name of the chat container"""
        return self._name

    @name.setter
    def name(self, name: str) -> None:
        """Set the name of the chat container"""
        name = name.strip()
        if self._name == name:
            return
        self._name = name
        self._changes.add("name")

        self.save()

    @property
    def system_prompt(self) -> ParllamaChatMessage | None:
        """Get the system message"""
        if len(self.messages) > 0 and self.messages[0].role == "system":
            return self.messages[0]
        return None

    # system prompt setter
    @system_prompt.setter
    def system_prompt(self, value: ParllamaChatMessage | None) -> None:
        """Set system prompt"""
        if not value:
            if len(self.messages) == 0 or self.messages[0].role != "system":
                return
            msg = self.messages.pop(0)
            self.last_updated = datetime.now(UTC)
            self._changes.add("messages")
            self._changes.add("system_prompt")
            self.save()
            if msg.parent:
                self.post_message(ParChatMessageDeleted(parent_id=msg.parent.id, message_id=msg.id))
            return

        if len(self.messages) > 0 and self.messages[0].role == "system":
            msg: ParllamaChatMessage = self.messages[0]
            if msg.content == value.content:
                return
            msg.content = value.content
            self.last_updated = datetime.now(UTC)
            self._changes.add("messages")
            self._changes.add("system_prompt")
            self.save()
        else:
            self.add_message(value, True)

    def get_first_user_message(self) -> ParllamaChatMessage | None:
        """Get the first user message"""
        for msg in self.messages:
            if msg.role == "user":
                return msg
        return None

    def get_last_user_message(self) -> ParllamaChatMessage | None:
        """Get the last user message"""
        if len(self.messages) == 0:
            return None
        msg = self.messages[-1]
        if msg.role == "user":
            return msg
        return None

    def __iter__(self):
        """Iterate over messages"""
        return iter(self.messages)

    def __len__(self) -> int:
        """Get the number of messages"""
        return len(self.messages)

    def __getitem__(self, msg_id: str) -> ParllamaChatMessage | None:
        """Get a message"""
        return self._id_to_msg.get(msg_id)

    def __setitem__(self, msg_id: str, value: ParllamaChatMessage) -> None:
        """Set a message"""
        for i, msg in enumerate(self.messages):
            if msg.id == msg_id:
                self._id_to_msg[msg_id] = value
                self.messages[i] = value
                self.last_updated = datetime.now(UTC)
                self._changes.add("messages")
                self.save()
                return
        self.add_message(value)

    def __delitem__(self, key: str) -> None:
        """Delete a message"""
        del self._id_to_msg[key]
        for i, msg in enumerate(self.messages):
            if msg.id == key:
                if msg.parent:
                    self.post_message(ParChatMessageDeleted(parent_id=msg.parent.id, message_id=msg.id))

                msg = self.messages.pop(i)
                self.last_updated = datetime.now(UTC)
                if msg.role == "system":
                    self._changes.add("system_prompt")
                self._changes.add("messages")
                self.save()
                return

    def __contains__(self, item: ParllamaChatMessage) -> bool:
        """Check if a message exists"""
        return item.id in self._id_to_msg

    def __iadd__(self, other: ParllamaChatMessage) -> ChatMessageContainer:
        """Add a message to the chat"""
        self.add_message(other)
        return self

    def __str__(self) -> str:
        """Get a string representation of the chat"""
        ret = StringIO()
        ret.write(f"# {self.name}\n\n")
        for msg in self.messages:
            ret.write(str(msg))
        return ret.getvalue()

    def to_json(self) -> str:
        """Convert the chat session to JSON"""
        return json.dumps(
            {
                "id": self.id,
                "name": self._name,
                "last_updated": self.last_updated.isoformat(),
                "messages": [m.to_dict() for m in self.messages],
            },
            str,
            json.OPT_INDENT_2,
        ).decode("utf-8")

    def export_as_markdown(self, filename: str) -> bool:
        """Save the chat session to markdown file"""
        try:
            with open(os.path.join(settings.export_md_dir, filename), "w", encoding="utf-8") as f:
                f.write(str(self))
            return True
        except OSError:
            return False

    @property
    def context_length(self) -> int:
        """Return current message context length"""
        total: int = 0
        for msg in self.messages:
            total += len(msg.content)
        return total

    def save(self) -> bool:
        """Save chats to file"""
        raise NotImplementedError("save not implemented in base class")

    def clear_messages(self) -> None:
        """Clear all messages"""
        self.messages.clear()
        self._id_to_msg.clear()
        self._changes.add("messages")

    def clear_changes(self) -> None:
        """Clear changes"""
        # self.log_it("Clearing changes")
        self._changes.clear()

    @property
    def is_dirty(self) -> bool:
        """Check if there are any changes"""
        # self.log_it(",".join(self._changes))
        return len(self._changes) > 0

    @is_dirty.setter
    def is_dirty(self, value: bool) -> None:
        """Set dirty status"""
        if value:
            self._changes.add("is_dirty")
        else:
            self.clear_changes()

    @contextmanager
    def batch_changes(self) -> Generator[None, None, None]:
        """Batch changes"""
        # self.log_it("Starting batch changes")
        self._batching = True
        yield
        self._batching = False
        # self.log_it("Committing batch changes")
        self.save()



"""Data classes for job queue"""

from __future__ import annotations

from uuid import UUID, uuid4

from pydantic import BaseModel, Field


class QueueJob(BaseModel):
    """Base job"""

    id: UUID = Field(default_factory=uuid4)
    modelName: str


class PushModelJob(QueueJob):
    """Push model job"""


class PullModelJob(QueueJob):
    """Pull model job"""


class DeleteModelJob(QueueJob):
    """Delete model job"""


class CopyModelJob(QueueJob):
    """Copy model job"""

    dstModelName: str


class CreateModelJob(QueueJob):
    """Create model job"""

    modelCode: str
    quantizationLevel: str | None



"""The main application class."""

from __future__ import annotations

import asyncio
from collections.abc import Iterator
from queue import Empty, Queue

import clipman as clipboard
import humanize
import ollama
from httpx import ConnectError
from ollama import ProgressResponse
from par_ai_core.llm_providers import LlmProvider
from rich.columns import Columns
from rich.console import ConsoleRenderable, RenderableType, RichCast
from rich.progress_bar import ProgressBar
from rich.style import Style
from rich.text import Text
from textual import on, work
from textual.app import App
from textual.binding import Binding
from textual.color import Color
from textual.message import Message
from textual.message_pump import MessagePump
from textual.notifications import SeverityLevel
from textual.timer import Timer
from textual.widget import Widget
from textual.widgets import Input, Select, TextArea

from parllama import __application_title__
from parllama.chat_manager import ChatManager, chat_manager
from parllama.dialogs.help_dialog import HelpDialog
from parllama.dialogs.information import InformationDialog
from parllama.dialogs.theme_dialog import ThemeDialog
from parllama.messages.messages import (
    ChangeTab,
    ClearChatInputHistory,
    DeletePrompt,
    DeleteSession,
    LocalCreateModelFromExistingRequested,
    LocalModelCopied,
    LocalModelCopyRequested,
    LocalModelCreated,
    LocalModelCreateRequested,
    LocalModelDelete,
    LocalModelDeleted,
    LocalModelListLoaded,
    LocalModelListRefreshRequested,
    LocalModelPulled,
    LocalModelPullRequested,
    LocalModelPushed,
    LocalModelPushRequested,
    LogIt,
    ModelInteractRequested,
    PromptListChanged,
    PromptListLoaded,
    PromptSelected,
    ProviderModelsChanged,
    PsMessage,
    RefreshProviderModelsRequested,
    RegisterForUpdates,
    SendToClipboard,
    SessionListChanged,
    SessionSelected,
    SessionToPrompt,
    SetModelNameLoading,
    SiteModelsLoaded,
    SiteModelsRefreshRequested,
    StatusMessage,
    UnRegisterForUpdates,
)
from parllama.models.jobs import CopyModelJob, CreateModelJob, PullModelJob, PushModelJob, QueueJob
from parllama.ollama_data_manager import ollama_dm
from parllama.prompt_utils.import_fabric import import_fabric_manager
from parllama.provider_manager import provider_manager
from parllama.screens.main_screen import MainScreen
from parllama.secrets_manager import secrets_manager
from parllama.settings_manager import settings
from parllama.theme_manager import theme_manager
from parllama.update_manager import update_manager


class ParLlamaApp(App[None]):
    """Main application class"""

    TITLE = __application_title__
    COMMAND_PALETTE_BINDING = "ctrl+underscore"
    ENABLE_COMMAND_PALETTE = False
    BINDINGS = [
        Binding(key="f1", action="help", description="Help", show=True, priority=True),
        Binding(key="ctrl+q", action="app.shutdown", description="Quit", show=True),
        Binding(
            key="f10",
            action="change_theme",
            description="Change Theme",
            show=True,
            priority=True,
        ),
        Binding(key="ctrl+c", action="noop", show=False),
    ]

    commands: list[dict[str, str]] = [
        {
            "action": "action_quit",
            "cmd": "Quit Application",
            "help": "Quit the application as soon as possible",
        }
    ]
    CSS_PATH = "app.tcss"

    notify_subs: dict[str, set[MessagePump]]
    main_screen: MainScreen
    job_queue: Queue[QueueJob]
    is_busy: bool = False
    last_status: RenderableType = ""
    chat_manager: ChatManager
    job_timer: Timer | None
    ps_timer: Timer | None

    def __init__(self) -> None:
        """Initialize the application."""
        super().__init__()
        self.notify_subs = {"*": set[MessagePump]()}
        theme_manager.set_app(self)
        provider_manager.set_app(self)
        secrets_manager.set_app(self)
        ollama_dm.set_app(self)
        chat_manager.set_app(self)
        update_manager.set_app(self)
        import_fabric_manager.set_app(self)

        self.job_timer = None
        self.ps_timer = None
        self.title = __application_title__

        self.job_queue = Queue[QueueJob]()
        self.is_busy = False
        self.is_refreshing = False
        self.last_status = ""
        if settings.theme_name not in theme_manager.list_themes():
            settings.theme_name = f"{settings.theme_name}_{settings.theme_mode}"
            if settings.theme_name not in theme_manager.list_themes():
                settings.theme_name = "par_dark"

        theme_manager.change_theme(settings.theme_name)

    async def on_mount(self) -> None:
        """Display the screen."""
        self.main_screen = MainScreen()

        await self.push_screen(self.main_screen)
        if settings.check_for_updates:
            await update_manager.check_for_updates()

        self.app.post_message(
            RegisterForUpdates(
                widget=self,
                event_names=[
                    "LocalModelPulled",
                    "LocalModelPushed",
                    "LocalModelCreated",
                    "LocalModelDeleted",
                    "LocalModelCopied",
                ],
            )
        )
        self.job_timer = self.set_timer(1, self.do_jobs)
        if settings.ollama_ps_poll_interval > 0:
            self.ps_timer = self.set_timer(1, self.update_ps)

        if settings.show_first_run:
            settings.show_first_run = False
            settings.save()
            self.set_timer(1, self.show_first_run)

        self.post_message(RefreshProviderModelsRequested(None))

    async def show_first_run(self) -> None:
        """Show first run screen"""
        await self.app.push_screen(
            InformationDialog(
                title="Welcome",
                message="""
Thank your for trying ParLlama!
Please take a moment to familiarize yourself with the the various options.
New options are being added all the time. Check out Whats New on the repo.
[link]https://github.com/paulrobello/parllama?tab=readme-ov-file#whats-new[/link]
By default ParLlama makes not attempt to connect to the internet.
If you would like to auto check for updates, you can enable it in the Startup section of Options.
                """,
            )
        )

    def action_noop(self) -> None:
        """Do nothing"""

    def action_help(self) -> None:
        """Show help screen"""
        self.app.push_screen(HelpDialog())

    def action_clear_field(self) -> None:
        """Clear focused widget value"""
        f: Widget | None = self.screen.focused
        if not f:
            return
        if isinstance(f, Input):
            f.value = ""
        if isinstance(f, TextArea):
            f.text = ""
        if isinstance(f, Select):
            f.value = Select.BLANK

    def action_copy_to_clipboard(self) -> None:
        """Copy focused widget value to clipboard"""
        f: Widget | None = self.screen.focused
        if not f:
            return

        if isinstance(f, Input | Select):
            self.app.post_message(SendToClipboard(str(f.value) if f.value and f.value != Select.BLANK else ""))

        if isinstance(f, TextArea):
            self.app.post_message(SendToClipboard(f.selected_text or f.text))

    def action_cut_to_clipboard(self) -> None:
        """Cut focused widget value to clipboard"""
        try:
            f: Widget | None = self.screen.focused
            if not f:
                return
            if isinstance(f, Input):
                clipboard.copy(f.value)
                f.value = ""
            if isinstance(f, Select):
                self.app.post_message(SendToClipboard(str(f.value) if f.value and f.value != Select.BLANK else ""))
            if isinstance(f, TextArea):
                clipboard.copy(f.selected_text or f.text)
                f.text = ""
        except Exception as _:
            self.notify("Error with clipboard", severity="error")

    @on(SendToClipboard)
    def send_to_clipboard(self, event: SendToClipboard) -> None:
        """Send string to clipboard"""
        # works for remote ssh sessions
        self.copy_to_clipboard(event.message)
        # works for local sessions
        try:
            clipboard.copy(event.message)
            if event.notify:
                self.notify("Copied to clipboard")
        except Exception as _:
            self.notify("Error with clipboard", severity="error")

    @on(LocalModelPushRequested)
    def on_model_push_requested(self, event: LocalModelPushRequested) -> None:
        """Push requested model event"""
        self.job_queue.put(PushModelJob(modelName=event.model_name))
        self.main_screen.local_view.post_message(SetModelNameLoading(event.model_name, True))
        # self.notify(f"Model push {msg.model_name} requested")

    @on(LocalModelCreateRequested)
    def on_model_create_requested(self, event: LocalModelCreateRequested) -> None:
        """Create model requested event"""
        self.job_queue.put(
            CreateModelJob(
                modelName=event.model_name,
                modelCode=event.model_code,
                quantizationLevel=event.quantization_level,
            )
        )

    @on(LocalModelDelete)
    def on_local_model_delete(self, event: LocalModelDelete) -> None:
        """Delete local model event"""
        if not ollama_dm.delete_model(event.model_name):
            self.main_screen.local_view.post_message(SetModelNameLoading(event.model_name, False))
            self.status_notify(f"Error deleting model {event.model_name}.", severity="error")
            return
        self.post_message_all(LocalModelDeleted(event.model_name))

    @on(LocalModelDeleted)
    def on_model_deleted(self, event: LocalModelDeleted) -> None:
        """Local model has been deleted event"""
        self.post_message_all(ProviderModelsChanged(provider=LlmProvider.OLLAMA))
        self.status_notify(f"Model {event.model_name} deleted.")

    @on(LocalModelPullRequested)
    def on_model_pull_requested(self, event: LocalModelPullRequested) -> None:
        """Pull requested model event"""
        if event.notify:
            self.notify(f"Model pull {event.model_name} queued")
        self.job_queue.put(PullModelJob(modelName=event.model_name))
        self.post_message_all(SetModelNameLoading(event.model_name, True))

    @on(LocalModelCopyRequested)
    def on_local_model_copy_requested(self, event: LocalModelCopyRequested) -> None:
        """Local model copy request event"""
        self.job_queue.put(CopyModelJob(modelName=event.src_model_name, dstModelName=event.dst_model_name))

    async def do_copy_local_model(self, event: CopyModelJob) -> None:
        """Copy local model"""
        ret = ollama_dm.copy_model(event.modelName, event.dstModelName)
        self.main_screen.local_view.post_message(
            LocalModelCopied(
                src_model_name=event.modelName,
                dst_model_name=event.dstModelName,
                success=ret["status"] == "success",
            )
        )

    @on(LocalModelCopied)
    def on_local_model_copied(self, event: LocalModelCopied) -> None:
        """Local model copied event"""
        if event.success:
            self.status_notify(f"Model {event.src_model_name} copied to {event.dst_model_name}")
        else:
            self.status_notify(
                f"Copying model {event.src_model_name} to {event.dst_model_name} failed",
                severity="error",
            )

    async def do_progress(self, job: QueueJob, res: Iterator[ProgressResponse]) -> str:
        """Update progress bar embedded in status bar"""
        try:
            last_status = ""
            for msg in res:
                if settings.shutting_down:
                    break

                last_status = msg.status or ""
                pb: ProgressBar | None = None
                percent = ""
                if msg.total and msg.completed:
                    percent = str(int(msg.completed / msg.total * 100)) + "%"
                    primary_style = Style(color=Color.parse(self.current_theme.primary).rich_color)
                    background_style = Style(color=Color.parse(self.current_theme.surface or "#111").rich_color)
                    pb = ProgressBar(
                        total=msg.total or 0,
                        completed=msg.completed or 0,
                        width=25,
                        style=background_style,
                        complete_style=primary_style,
                        finished_style=primary_style,
                    )
                else:
                    percent = ""
                if percent and msg.status == "success":
                    percent = "100%"
                parts: list[RenderableType] = [
                    Text.assemble(
                        job.modelName,
                        " ",
                        msg.status or "",
                        " ",
                        percent,
                        " ",
                    )
                ]  # type: ignore
                if pb:
                    parts.append(pb)

                self.post_message_all(StatusMessage(Columns(parts), log_it=False))
            return last_status
        except ollama.ResponseError as e:
            self.post_message_all(StatusMessage(Text.assemble(("error:" + str(e), "red"))))
            raise e

    async def do_pull(self, job: PullModelJob) -> None:
        """Pull a model from ollama.com"""
        try:
            res: Iterator[ProgressResponse] = ollama_dm.pull_model(job.modelName)
            last_status = await self.do_progress(job, res)

            self.post_message_all(LocalModelPulled(model_name=job.modelName, success=last_status == "success"))
        except ollama.ResponseError as e:
            self.log_it(e)
            self.post_message_all(LocalModelPulled(model_name=job.modelName, success=False))

    async def do_push(self, job: PushModelJob) -> None:
        """Push a model to ollama.com"""
        try:
            res: Iterator[ProgressResponse] = ollama_dm.push_model(job.modelName)
            last_status = await self.do_progress(job, res)

            self.post_message_all(LocalModelPushed(model_name=job.modelName, success=last_status == "success"))
        except ollama.ResponseError:
            self.post_message_all(LocalModelPushed(model_name=job.modelName, success=False))

    async def do_create_model(self, job: CreateModelJob) -> None:
        """Create a new local model"""
        try:
            self.main_screen.log_view.richlog.write(job.modelCode)
            res = ollama_dm.create_model(job.modelName, job.modelCode, job.quantizationLevel)
            last_status = await self.do_progress(job, res)

            self.main_screen.local_view.post_message(
                LocalModelCreated(
                    model_name=job.modelName,
                    model_code=job.modelCode,
                    quantization_level=job.quantizationLevel,
                    success=last_status == "success",
                )
            )
        except ollama.ResponseError:
            self.main_screen.local_view.post_message(
                LocalModelCreated(
                    model_name=job.modelName,
                    model_code=job.modelCode,
                    quantization_level=job.quantizationLevel,
                    success=False,
                )
            )

    @work(group="do_jobs", thread=True)
    async def do_jobs(self) -> None:
        """poll for queued jobs"""
        while True:
            try:
                job: QueueJob = self.job_queue.get(block=True, timeout=1)
                if settings.shutting_down:
                    return
                if not job:
                    continue
                self.is_busy = True
                if isinstance(job, PullModelJob):
                    await self.do_pull(job)
                elif isinstance(job, PushModelJob):
                    await self.do_push(job)
                elif isinstance(job, CopyModelJob):
                    await self.do_copy_local_model(job)
                elif isinstance(job, CreateModelJob):
                    await self.do_create_model(job)
                else:
                    self.status_notify(
                        f"Unknown job type {type(job)}",
                        severity="error",
                    )
            except Empty:
                if self._exit:
                    return
                # TODO figure this out
                # if self.is_busy:
                #     self.post_message(LocalModelListRefreshRequested(widget=None))
                #     self.is_busy = False
                continue

    @on(LocalModelPulled)
    def on_model_pulled(self, event: LocalModelPulled) -> None:
        """Model pulled event"""
        if event.success:
            self.status_notify(
                f"Model {event.model_name} pulled.",
            )
        else:
            self.status_notify(
                f"Model {event.model_name} failed to pull.",
                severity="error",
            )

    @on(LocalModelCreated)
    def on_model_created(self, event: LocalModelCreated) -> None:
        """Model created event"""
        if event.success:
            self.status_notify(
                f"Model {event.model_name} created.",
            )
            self.set_timer(1, self.action_refresh_models)
        else:
            self.status_notify(
                f"Model {event.model_name} failed to create.",
                severity="error",
            )

    def action_refresh_models(self) -> None:
        """Refresh models action."""
        self.refresh_models()

    @on(UnRegisterForUpdates)
    def on_unregister_for_updates(self, event: UnRegisterForUpdates) -> None:
        """Unregister widget from all updates"""
        if not event.widget:
            return
        for _, s in self.notify_subs.items():
            s.discard(event.widget)

    @on(RegisterForUpdates)
    def on_register_for_updates(self, event: RegisterForUpdates) -> None:
        """Register for updates event"""
        for event_name in event.event_names:
            if event_name not in self.notify_subs:
                self.notify_subs[event_name] = set()
            self.notify_subs[event_name].add(event.widget)

    @on(LocalModelListRefreshRequested)
    def on_model_list_refresh_requested(self) -> None:
        """Model refresh request event"""
        if self.is_refreshing:
            self.status_notify("A model refresh is already in progress. Please wait.")
            return
        self.refresh_models()

    @work(group="refresh_models", thread=True)
    async def refresh_models(self):
        """Refresh the models."""
        self.is_refreshing = True
        try:
            self.post_message_all(StatusMessage("Local model list refreshing..."))
            ollama_dm.refresh_models()
            self.post_message_all(StatusMessage("Local model list refreshed"))
            self.post_message_all(LocalModelListLoaded())
            self.post_message_all(ProviderModelsChanged(provider=LlmProvider.OLLAMA))
        except ConnectError as e:
            self.post_message(
                LogIt(
                    f"Failed to refresh local models: {e}",
                    severity="error",
                    notify=True,
                )
            )
        finally:
            self.is_refreshing = False

    # @on(LocalModelListLoaded)
    # def on_model_data_loaded(self) -> None:
    #     """Refresh model completed"""
    #     self.post_message_all(StatusMessage("Local model list refreshed"))
    #     # self.notify("Local models refreshed.")

    @on(SiteModelsRefreshRequested)
    def on_site_models_refresh_requested(self, msg: SiteModelsRefreshRequested) -> None:
        """Site model refresh request event"""
        if self.is_refreshing:
            self.status_notify("A model refresh is already in progress. Please wait.")
            return
        self.refresh_site_models(msg)

    @on(SiteModelsLoaded)
    def on_site_models_loaded(self) -> None:
        """Site model refresh completed"""
        self.status_notify("Site models refreshed")

    @work(group="refresh_site_model", thread=True)
    async def refresh_site_models(self, msg: SiteModelsRefreshRequested):
        """Refresh the site model."""
        self.is_refreshing = True
        try:
            self.post_message_all(
                StatusMessage(f"Site models for {msg.ollama_namespace or 'models'} refreshing... force={msg.force}")
            )
            ollama_dm.refresh_site_models(msg.ollama_namespace, None, msg.force)
            self.main_screen.site_view.post_message(SiteModelsLoaded(ollama_namespace=msg.ollama_namespace))
            self.post_message_all(
                StatusMessage(f"Site models for {msg.ollama_namespace or 'models'} loaded. force={msg.force}")
            )

        finally:
            self.is_refreshing = False

    @work(group="update_ps", thread=True)
    async def update_ps(self) -> None:
        """Update ps status bar msg"""
        was_blank = False
        while not settings.shutting_down:
            if settings.ollama_ps_poll_interval < 1:
                self.post_message_all(PsMessage(msg=""))
                break
            await asyncio.sleep(settings.ollama_ps_poll_interval)
            ret = ollama_dm.model_ps()
            if len(ret.models) < 1:
                if not was_blank:
                    self.post_message_all(PsMessage(msg=""))
                was_blank = True
                continue
            was_blank = False
            info = ret.models[0]  # only take first one since ps status bar is a single line
            self.post_message_all(
                PsMessage(
                    msg=Text.assemble(
                        "Name: ",
                        info.name,
                        " Size: ",
                        humanize.naturalsize(info.size_vram),
                        " Processor: ",
                        ret.processor,
                        " Until: ",
                        humanize.naturaltime(info.expires_at),
                    )
                )
            )

    def status_notify(self, msg: str, severity: SeverityLevel = "information") -> None:
        """Show notification and update status bar"""
        self.notify(msg, severity=severity, timeout=5 if severity != "information" else 3)
        self.main_screen.post_message(StatusMessage(msg))

    def post_message_all(self, event: Message) -> None:
        """Post a message to all screens"""
        if isinstance(event, StatusMessage):
            if event.log_it:
                self.log(event.msg)
            self.last_status = event.msg
            self.main_screen.post_message(event)
            return
        if isinstance(event, PsMessage):
            self.main_screen.post_message(event)
            return
        sub_name = event.__class__.__name__
        if sub_name in self.notify_subs:
            for w in list(self.notify_subs[sub_name]):
                w.post_message(event)

    @on(ChangeTab)
    def on_change_tab(self, event: ChangeTab) -> None:
        """Change tab event"""
        event.stop()
        self.main_screen.change_tab(event.tab)

    @on(LocalCreateModelFromExistingRequested)
    def on_create_model_from_existing_requested(self, msg: LocalCreateModelFromExistingRequested) -> None:
        """Create model from existing event"""
        self.main_screen.create_view.name_input.value = f"my-{msg.model_name}"
        if not self.main_screen.create_view.name_input.value.endswith(":latest"):
            self.main_screen.create_view.name_input.value += ":latest"
        self.main_screen.create_view.text_area.text = msg.model_code
        self.main_screen.create_view.quantize_input.value = msg.quantization_level or ""
        self.main_screen.change_tab("Create")
        self.main_screen.create_view.name_input.focus()

    @on(ModelInteractRequested)
    async def on_model_interact_requested(self, event: ModelInteractRequested) -> None:
        """Model interact requested event"""
        await self.main_screen.chat_view.action_new_tab()
        self.main_screen.chat_view.active_tab.session_config.provider_model_select.provider_select.value = (
            LlmProvider.OLLAMA
        )
        self.main_screen.chat_view.active_tab.session_config.provider_model_select.model_select.deferred_value = (
            event.model_name
        )
        self.main_screen.change_tab("Chat")
        self.main_screen.chat_view.user_input.focus()

    @on(ClearChatInputHistory)
    def on_clear_chat_history(self, event: ClearChatInputHistory) -> None:
        """Clear chat history event"""
        event.stop()
        self.post_message_all(event)

    @on(SessionListChanged)
    def on_session_list_changed(self, event: SessionListChanged) -> None:
        """Session list changed event"""
        event.stop()
        self.post_message_all(event)

    @on(PromptListChanged)
    def on_prompt_list_changed(self, event: PromptListChanged) -> None:
        """Prompt list changed event"""
        event.stop()
        self.post_message_all(event)

    @on(SessionSelected)
    def on_session_selected(self, event: SessionSelected) -> None:
        """Session selected event"""
        event.stop()
        self.post_message_all(event)

    @on(PromptSelected)
    def on_prompt_selected(self, event: PromptSelected) -> None:
        """Session selected event"""
        event.stop()
        self.post_message_all(event)

    @on(DeleteSession)
    def on_delete_session(self, event: DeleteSession) -> None:
        """Delete session event"""
        event.stop()
        self.post_message_all(event)

    @on(DeletePrompt)
    def on_delete_prompt(self, event: DeletePrompt) -> None:
        """Delete prompt event"""
        event.stop()
        self.post_message_all(event)

    @on(SessionToPrompt)
    def on_session_to_prompt(self, event: SessionToPrompt) -> None:
        """Session to prompt event"""
        event.stop()
        chat_manager.session_to_prompt(event.session_id, event.submit_on_load, event.prompt_name)

    @on(PromptListLoaded)
    def on_prompt_list_loaded(self, event: PromptListLoaded) -> None:
        """Prompt list loaded event"""
        event.stop()
        self.post_message_all(event)

    @on(LogIt)
    def on_log_it(self, event: LogIt) -> None:
        """Log an event to the log view"""
        event.stop()
        self.log_it(event.msg)
        if event.notify and isinstance(event.msg, str):
            self.notify(
                event.msg,
                severity=event.severity,
                timeout=event.timeout or 5,
            )

    def log_it(self, msg: ConsoleRenderable | RichCast | str | object) -> None:
        """Log a message to the log view"""
        self.main_screen.log_view.richlog.write(msg)

    async def action_shutdown(self) -> None:
        """Quit the application"""
        settings.shutting_down = True
        await self.action_quit()

    @work(exclusive=True)
    async def refresh_provider_models(self) -> None:
        """Refresh provider models"""
        provider_manager.refresh_models()

    @on(RefreshProviderModelsRequested)
    def on_refresh_provider_models(self, event: RefreshProviderModelsRequested) -> None:
        """Refresh provider models event"""
        event.stop()
        self.refresh_provider_models()

    @on(ProviderModelsChanged)
    def on_provider_models_refreshed(self, event: ProviderModelsChanged) -> None:
        """Provider models refreshed event"""
        event.stop()
        self.post_message_all(event)

    @work
    async def action_change_theme(self) -> None:
        """An action to change the theme."""
        theme = await self.push_screen_wait(ThemeDialog())
        settings.theme_name = theme



"""Input widget with special tab completion and history."""

from __future__ import annotations

from textual import events
from textual.binding import Binding

from parllama.messages.messages import HistoryNext, HistoryPrev, ToggleInputMode
from parllama.widgets.input_tab_complete import InputTabComplete


class InputWithHistory(InputTabComplete):
    """Input widget with special tab completion and history."""

    BINDINGS = [
        Binding(key="ctrl+j", action="toggle_mode", description="Multi Line", show=True),
    ]

    def __init__(
        self,
        **kwargs,
    ) -> None:
        """Initialize the Input."""
        super().__init__(**kwargs)

    async def _on_key(self, event: events.Key) -> None:
        """Override tab, up and down key behavior."""

        if event.key == "tab":
            self._cursor_visible = True
            if self.cursor_blink and self._blink_timer:
                self._blink_timer.reset()
            if self._cursor_at_end and self._suggestion and self.value != self._suggestion:
                self.value = self._suggestion
                self.cursor_position = len(self.value)
                if self.submit_on_complete:
                    await self.action_submit()
                else:
                    event.stop()
                    event.prevent_default()
            else:
                if self.submit_on_tab:
                    await self.action_submit()

        if event.key in ("up", "down"):
            event.stop()
            event.prevent_default()
            if event.key == "up":
                self.post_message(HistoryPrev(input=self))
            if event.key == "down":
                self.post_message(HistoryNext(input=self))
            return
        return await super()._on_key(event)

    def action_toggle_mode(self) -> None:
        """Request input mode toggle"""
        self.post_message(ToggleInputMode())



"""Prompt messages for par event system"""

from __future__ import annotations

from dataclasses import dataclass

from parllama.messages.shared import PromptChanges
from parllama.par_event_system import ParEventBase


@dataclass
class ParPromptMessage(ParEventBase):
    """Session message base class"""

    prompt_id: str


@dataclass
class ParPromptUpdated(ParPromptMessage):
    """Session was updated"""

    changed: PromptChanges


@dataclass
class ParPromptDelete(ParPromptMessage):
    """Request prompt be deleted."""


@dataclass
class ParPromptChatMessage(ParPromptMessage):
    """Chat message base class"""

    message_id: str


@dataclass
class ParPromptChatUpdated(ParPromptChatMessage):
    """Chat message updated"""



"""Theme manager for Textual"""

from __future__ import annotations

import os
import shutil
from pathlib import Path
from typing import Any, Literal, TypeAlias

import orjson as json
from textual.app import App
from textual.theme import Theme

from parllama.par_event_system import ParEventSystemBase
from parllama.settings_manager import settings

ThemeMode: TypeAlias = Literal["dark", "light"]
ThemeModes: list[ThemeMode] = ["dark", "light"]
Themes: TypeAlias = dict[str, Theme]


class InvalidThemeError(Exception):
    """Raised when an invalid theme is provided."""

    def __init__(self, theme_name: str):
        """Initialize the exception with the invalid theme name."""
        self.theme_name = theme_name
        super().__init__(f"Invalid theme: {theme_name}")


class ThemeModeError(InvalidThemeError):
    """Raised when a theme does not have at least one of 'dark' or 'light' modes."""

    def __init__(self, theme_name: str):
        """Initialize the exception with the invalid theme name."""
        super().__init__(f"Theme '{theme_name}' does not have at least one of 'dark' or 'light' modes.")


class ThemeManager(ParEventSystemBase):
    """Theme manager for Textual"""

    theme_folder: Path

    def __init__(self) -> None:
        """Initialize the theme manager"""
        super().__init__(id="theme_manager")
        self.theme_folder = Path(settings.data_dir) / "themes"

    def set_app(self, app: App[Any] | None) -> None:
        """Set the app and load existing sessions and prompts from storage"""
        super().set_app(app)
        self.load_themes()

    def ensure_default_theme(self) -> None:
        """Ensure that the default theme exists."""
        default_theme_file: Path = Path(os.path.dirname(Path(__file__))) / "themes" / "par.json"

        if not self.theme_folder.exists():
            self.theme_folder.mkdir(parents=True, exist_ok=True)
        theme_file: Path = self.theme_folder / "par.json"

        if not theme_file.exists():
            shutil.copy(default_theme_file, theme_file)

    def load_theme(self, theme_name: str) -> None:
        """Load textual theme from json file"""
        if not self.app:
            raise Exception("App is not initialized")

        theme_name = os.path.basename(theme_name)
        theme_file: Path = Path(self.theme_folder) / (theme_name + ".json")
        theme_def = json.loads(theme_file.read_bytes())
        if "dark" not in theme_def and "light" not in theme_def:
            raise ThemeModeError(theme_name)

        for mode in ThemeModes:
            if mode in theme_def:
                self.app.register_theme(
                    Theme(
                        name=f"{theme_name}_{mode}",
                        **theme_def[mode],
                    )
                )

    def load_themes(self) -> None:
        """Load textual themes from json files"""
        self.ensure_default_theme()

        for file in os.listdir(self.theme_folder):
            if file.lower().endswith(".json"):
                self.load_theme(f"{self.theme_folder}/{os.path.splitext(file)[0]}")

    def get_theme(self, theme_name: str) -> Theme:
        """Get theme by name"""
        if not self.app:
            raise Exception("App is not initialized")
        return self.app.available_themes[theme_name]

    def list_themes(self) -> list[str]:
        """Get list of themes"""
        if not self.app:
            raise Exception("App is not initialized")

        return list(self.app.available_themes.keys())

    def theme_select_options(self) -> list[tuple[str, str]]:
        """Get select options for theme"""
        return [(theme, theme) for theme in self.list_themes()]

    def change_theme(self, theme_name: str) -> None:
        """Change the theme"""
        if not self.app:
            raise Exception("App is not initialized")
        self.app.theme = theme_name


theme_manager = ThemeManager()



"""Textarea widget with special tab completion and history."""

from __future__ import annotations

from dataclasses import dataclass

from textual import events
from textual.binding import Binding
from textual.message import Message
from textual.widgets import TextArea

from parllama.messages.messages import HistoryNext, HistoryPrev, ToggleInputMode


class UserTextArea(TextArea):
    """Input widget with special tab completion."""

    @dataclass
    class Submitted(Message):
        """Posted when the enter key is pressed within an `Input`.

        Can be handled using `on_input_submitted` in a subclass of `Input` or in a
        parent widget in the DOM.
        """

        input: UserTextArea
        """The `Input` widget that is being submitted."""
        value: str
        """The value of the `Input` being submitted."""

        @property
        def control(self) -> UserTextArea:
            """Alias for self.input."""
            return self.input

    BINDINGS = [
        Binding(key="ctrl+g", action="submit", description="Submit", show=True),
        Binding(key="ctrl+j", action="toggle_mode", description="Single Line", show=True),
    ]

    def __init__(
        self,
        **kwargs,
    ) -> None:
        """Initialize the Input."""
        super().__init__(tab_behavior="indent", **kwargs)
        self._hit_boundary = False

    async def _on_key(self, event: events.Key) -> None:
        """Override tab, up and down key behavior."""
        self._restart_blink()
        if event.key in ("up", "down"):
            if event.key == "up" and self.cursor_at_first_line:
                if self._hit_boundary:
                    event.stop()
                    event.prevent_default()

                    self.post_message(HistoryPrev(input=self))
                    return
                self._hit_boundary = True
            if event.key == "down" and self.cursor_at_last_line:
                if self._hit_boundary:
                    event.stop()
                    event.prevent_default()

                    self.post_message(HistoryNext(input=self))
                    return
                self._hit_boundary = True
            return await super()._on_key(event)

        self._hit_boundary = False
        if self.read_only:
            return
        return await super()._on_key(event)

    def action_submit(self) -> None:
        """Store the last input in history."""
        v: str = self.text.strip()
        if not v:
            return
        self.post_message(UserTextArea.Submitted(self, value=v))

    def action_toggle_mode(self) -> None:
        """Request input mode toggle"""
        self.post_message(ToggleInputMode())



"""Prompt manager class"""

from __future__ import annotations

import os
import uuid
from dataclasses import dataclass
from datetime import UTC, datetime
from pathlib import Path

import orjson as json
import pytz
import rich.repr

from parllama.chat_message import ParllamaChatMessage
from parllama.chat_message_container import ChatMessageContainer
from parllama.messages.par_prompt_messages import ParPromptDelete, ParPromptUpdated
from parllama.messages.shared import PromptChanges, prompt_change_list
from parllama.settings_manager import settings


@rich.repr.auto
@dataclass
class ChatPrompt(ChatMessageContainer):
    """Chat prompt class"""

    _description: str
    messages: list[ParllamaChatMessage]
    last_updated: datetime
    source: str | None

    _submit_on_load: bool
    _id_to_msg: dict[str, ParllamaChatMessage]

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        *,
        id: str | None = None,  # pylint: disable=redefined-builtin
        name: str,
        description: str,
        messages: list[ParllamaChatMessage] | list[dict] | None = None,
        submit_on_load: bool = False,
        last_updated: datetime | None = None,
        source: str | None = None,
    ):
        """Initialize the chat prompt"""
        super().__init__(id=id, name=name, messages=messages, last_updated=last_updated)
        self._description = description
        self._submit_on_load = submit_on_load
        self.source = source

    def load(self) -> None:
        """Load chat prompts from files"""
        if self._loaded:
            return
        self._batching = True

        file_path = Path(settings.prompt_dir) / f"{self.id}.json"
        if not file_path.exists():
            return

        try:
            data: dict = json.loads(file_path.read_bytes())
            self.clear_messages()
            msgs = data["messages"] or []
            for m in msgs:
                self.add_message(ParllamaChatMessage(**m))
            self._loaded = True
        except Exception as e:  # pylint: disable=broad-exception-caught
            self.log_it(f"Error loading prompt {e}", notify=True, severity="error")
        finally:
            self._batching = False

    def delete(self) -> None:
        """Delete the prompt"""
        self.post_message(ParPromptDelete(prompt_id=self.id))

    def _notify_changed(self, changed: PromptChanges) -> None:
        """Notify changed"""
        self.post_message(ParPromptUpdated(prompt_id=self.id, changed=changed))

    @property
    def description(self) -> str:
        """Get the description of the chat prompt"""
        return self._description

    @description.setter
    def description(self, value: str) -> None:
        """Set the description of the chat prompt"""
        value = value.strip()
        if self._description == value:
            return
        self._description = value
        self._changes.add("description")
        self.save()

    @property
    def submit_on_load(self) -> bool:
        """Get whether the prompt should be submitted on load"""
        return self._submit_on_load

    @submit_on_load.setter
    def submit_on_load(self, value: bool) -> None:
        """Set whether the prompt should be submitted on load"""
        if self._submit_on_load == value:
            return
        self._submit_on_load = value
        self._changes.add("submit_on_load")
        self.save()

    def new_prompt(self, prompt_name: str = "My Prompt"):
        """Start new session"""
        self.id = uuid.uuid4().hex
        self._name = prompt_name
        self._description = ""
        self.source = ""
        self.messages.clear()
        self._id_to_msg.clear()
        self._loaded = False
        self.batching = False
        self.clear_changes()

    def __eq__(self, other: object) -> bool:
        """Check if two sessions are equal"""
        if not isinstance(other, ChatPrompt):
            return NotImplemented
        return self.id == other.id

    def __ne__(self, other: object) -> bool:
        """Check if two sessions are not equal"""
        if not isinstance(other, ChatPrompt):
            return NotImplemented
        return self.id != other.id

    def to_json(self) -> str:
        """Convert the chat session to JSON"""
        return json.dumps(
            {
                "id": self.id,
                "name": self.name,
                "last_updated": self.last_updated.isoformat(),
                "description": self._description,
                "submit_on_load": self._submit_on_load,
                "messages": [m.to_dict() for m in self.messages],
                "source": self.source,
            },
            str,
            json.OPT_INDENT_2,
        ).decode("utf-8")

    @staticmethod
    def from_json(json_data: str, load_messages: bool = False) -> ChatPrompt:
        """Convert JSON to chat session"""
        data: dict = json.loads(json_data)
        utc = pytz.UTC

        return ChatPrompt(
            id=data["id"],
            name=data["name"],
            last_updated=datetime.fromisoformat(data["last_updated"]).replace(tzinfo=utc),
            description=data.get("description", ""),
            messages=([ParllamaChatMessage(**m) for m in data["messages"]] if load_messages else None),
            submit_on_load=data.get("submit_on_load", False),
            source=data.get("source"),
        )

    @staticmethod
    def load_from_file(filename: str) -> ChatPrompt | None:
        """Load a chat prompt from a file"""
        try:
            with open(os.path.join(settings.prompt_dir, filename), encoding="utf-8") as f:
                return ChatPrompt.from_json(f.read())
        except OSError:
            return None

    @property
    def is_valid(self) -> bool:
        """return true if session is valid"""
        return len(self.name) > 0

    def save(self) -> bool:
        """Save the chat prompt to a file"""
        if self._batching:
            # self.log_it(f"CP is batching, not notifying: {self.name}")
            return False
        if not self._loaded:
            self.load()
        if not self.is_dirty:
            # self.log_it(f"CP is not dirty, not notifying: {self.name}")
            return False  # No need to save if no changes
        self.last_updated = datetime.now(UTC)
        nc: PromptChanges = PromptChanges()
        for change in self._changes:
            if change in prompt_change_list:
                nc.add(change)  # type: ignore

        self._notify_changed(nc)
        self.clear_changes()

        if not self.is_valid:
            # self.log_it(f"CP not valid, not saving: {self.id}")
            return False  # Cannot save without name

        file_name = f"{self.id}.json"  # Use prompt ID as filename
        try:
            with open(os.path.join(settings.prompt_dir, file_name), "w", encoding="utf-8") as f:
                f.write(self.to_json())
            return True
        except OSError:
            return False

    def replace_messages(self, new_messages: list[ParllamaChatMessage]) -> None:
        """Replace all messages with new ones"""
        self.messages = new_messages
        self._id_to_msg = {m.id: m for m in new_messages}
        self._changes.add("messages")

    def clone(self, new_id: bool = False) -> ChatPrompt:
        """Clone the chat prompt"""
        return ChatPrompt(
            id=uuid.uuid4().hex if new_id else self.id,
            name=self.name,
            description=self._description,
            messages=[m.clone() for m in self.messages],
            submit_on_load=self._submit_on_load,
            last_updated=self.last_updated,
            source=self.source,
        )



"""View for the model tools."""

from __future__ import annotations

import os

from textual import on
from textual.app import ComposeResult
from textual.containers import Container, Vertical, VerticalScroll
from textual.events import Show
from textual.widgets import Button, ContentSwitcher, Static

from parllama.messages.messages import ChangeTab
from parllama.widgets.clickable_label import CopyToClipboardLabel


class ModelToolsView(Container):
    """View for the model tools."""

    DEFAULT_CSS = """
    ModelToolsView {
      #publish_panel {
        padding: 1;
        border: solid $primary;
        height: auto;
      }
      #pub_key {
        border: solid $primary;
        height: 4;
      }
    }
    """

    BINDINGS = []

    def __init__(self, **kwargs) -> None:
        """Initialise the screen."""
        super().__init__(**kwargs)

    def compose(self) -> ComposeResult:
        """Compose the content of the view."""
        pub_key: str = ""
        pub_key_path = os.path.join(os.path.expanduser("~"), ".ollama", "id_ed25519.pub")
        if os.path.exists(pub_key_path):
            with open(pub_key_path, encoding="utf-8") as f:
                pub_key = f.read().strip()
        with ContentSwitcher(initial="menu"):
            with VerticalScroll(id="menu"):
                yield Button("Create new model", id="new_model", variant="success")
                with Vertical(id="publish_panel") as v:
                    v.border_title = "Setup Ollama for pushing to your namespace"
                    with Vertical(id="pub_key") as v:
                        v.border_title = "This machines Ollama public key"
                        yield Static("Click the key below to copy to clipboard:")
                        yield CopyToClipboardLabel(pub_key)
                    yield Static(
                        "[@click=screen.open_keys_page]Open https://ollama.com/settings/keys[/]",
                        id="open_keys_page",
                    )

    def _on_show(self, event: Show) -> None:
        """Handle show event"""
        self.screen.sub_title = (  # pylint: disable=attribute-defined-outside-init
            "Model tools"
        )

    @on(Button.Pressed, "#new_model")
    def action_new_model(self) -> None:
        """Open the new model screen."""
        self.app.post_message(ChangeTab(tab="Create"))

    def action_open_keys_page(self):
        """Open the Ollama keys page."""
        self.app.open_url("https://ollama.com/settings/keys")
        self.notify("Ollama keys page opened")



"""Field set consisting of Label and Input for editing items."""

from __future__ import annotations

from par_ai_core.utils import to_class_case
from textual import on
from textual.app import ComposeResult
from textual.containers import Horizontal
from textual.widget import Widget
from textual.widgets import Button, Input, Label, Select, TextArea

from parllama.icons import COPY_EMOJI
from parllama.messages.messages import SendToClipboard
from parllama.widgets.hidden_input import HiddenInputField


class FieldSet(Widget):
    """Field set consisting of Label and Input for editing items."""

    DEFAULT_CSS = """
    FieldSet {
        width: 1fr;
        height: 1;
        Label {
            width: 9;
            margin-right: 1;
        }
        &> Horizontal {
            width: 1fr;
            &> Horizontal {
                width: 3fr;
            }
        }
        Input, Select, TextArea, Checkbox, HiddenInputField{
            width: 1fr;
        }
    }
    """
    label: Label
    """Label widget."""
    input: Widget
    """Input widget."""

    show_copy_button: bool = False
    """Show copy button to right of input."""
    _extra_children: list[Widget] | None = None

    def __init__(
        self,
        label: str,
        input_widget: Widget,
        *,
        show_copy_button: bool = False,
        extra_children: list[Widget] | None = None,
    ) -> None:
        """Initialize the field set."""
        base_name = to_class_case(label.replace("/", ""))
        super().__init__(id=f"{base_name}FieldSet", name=f"{base_name}FieldSet", classes="field_set")
        self.show_copy_button = show_copy_button
        self._extra_children = extra_children
        self.input = input_widget
        self.label = Label(label, id=f"{base_name}Label")

    def compose(self) -> ComposeResult:
        """Compose the field set."""
        with Horizontal():
            yield self.label
            with Horizontal():
                yield self.input
                if self.show_copy_button:
                    btn = Button(COPY_EMOJI, id="copy", classes="field-button")
                    btn.tooltip = "Copy to clipboard"
                    yield btn
                if self._extra_children:
                    yield from self._extra_children

    @on(Button.Pressed, "#copy")
    def copy(self, event: Button.Pressed) -> None:
        """Copy field value."""
        event.stop()

        if isinstance(self.input, Input | Select):
            self.app.post_message(SendToClipboard(str(self.input.value) if self.input.value else ""))
        elif isinstance(self.input, TextArea):
            self.app.post_message(SendToClipboard(self.input.text))
        elif isinstance(self.input, HiddenInputField):
            self.app.post_message(SendToClipboard(self.input.value))



"""Session manager class"""

from __future__ import annotations

import ast
import base64
import os
import uuid
from collections.abc import Iterator
from dataclasses import dataclass
from datetime import UTC, datetime
from pathlib import Path

import orjson as json
import pytz
import rich.repr
from langchain_core.messages import BaseMessageChunk
from par_ai_core.llm_config import LlmConfig, llm_run_manager
from par_ai_core.llm_providers import LlmProvider
from textual.message_pump import MessagePump

from parllama.chat_message import ParllamaChatMessage
from parllama.chat_message_container import ChatMessageContainer
from parllama.messages.messages import (
    ChatGenerationAborted,
    ChatMessage,
    ChatMessageDeleted,
    SessionChanges,
    SessionMessage,
    SessionUpdated,
)
from parllama.messages.par_chat_messages import ParChatMessageDeleted, ParChatUpdated
from parllama.messages.par_session_messages import ParSessionAutoName, ParSessionDelete, ParSessionUpdated
from parllama.messages.shared import session_change_list
from parllama.models.token_stats import TokenStats
from parllama.settings_manager import settings


@rich.repr.auto
@dataclass
class ChatSession(ChatMessageContainer):
    """Chat session class"""

    name_generated: bool
    """Set to True if the session name has been generated by LLM"""
    _subs: set[MessagePump]
    _abort: bool
    """Set to True if the session should be aborted"""
    _generating: bool
    """Set to True if the session is currently generating"""
    _batching: bool
    """Set to True if the session is currently being batched"""
    _stream_stats: TokenStats | None
    """Stream statistics"""
    _llm_config: LlmConfig
    """LLM configuration"""
    _key_secure: str | None
    """Encrypted current session password"""
    _key: bytes | None
    """Decryption key derived from the password."""
    _salt: bytes | None
    """Used with password to derive the en/decryption key."""

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        *,
        id: str | None = None,  # pylint: disable=redefined-builtin
        name: str,
        llm_config: LlmConfig,
        messages: list[ParllamaChatMessage] | list[dict] | None = None,
        last_updated: datetime | None = None,
    ):
        """Initialize the chat session"""
        super().__init__(id=id, name=name, messages=messages, last_updated=last_updated)
        self._batching = True
        self._salt = None
        self._key = None
        self._key_secure = None
        self.name_generated = False
        self._abort = False
        self._generating = False
        self._subs = set()
        self._stream_stats = None
        self._batching = False
        self._llm_config = llm_config

    @property
    def llm_config(self) -> LlmConfig:
        """Return the LLM configuration"""
        return self._llm_config

    def load(self) -> None:
        """Load chat sessions from files"""
        if self._loaded:
            self._batching = False
            return

        self._batching = True
        try:
            self._stream_stats = None
            file_path = Path(settings.chat_dir) / f"{self.id}.json"
            if not file_path.exists():
                return

            data: dict = json.loads(file_path.read_bytes())

            msgs = data["messages"] or []
            for m in msgs:
                if "message_id" in m:
                    m["id"] = "message_id"
                    del m["message_id"]
            for m in msgs:
                self.add_message(ParllamaChatMessage(**m))
            self._loaded = True
        except Exception as e:  # pylint: disable=broad-exception-caught
            self.log_it(f"Error loading session {e}", notify=True, severity="error")
        finally:
            self._batching = False
            self.clear_changes()

    def add_sub(self, sub: MessagePump) -> None:
        """Add a subscription"""
        self._subs.add(sub)

    def remove_sub(self, sub: MessagePump) -> None:
        """Remove a subscription"""
        self._subs.discard(sub)
        if self.num_subs == 0 and not self.is_valid:
            self.delete()

    def delete(self) -> None:
        """Delete the session"""
        self.post_message(ParSessionDelete(session_id=self.id))

    def _notify_subs(self, event: SessionMessage | ChatMessage | ChatMessageDeleted) -> None:
        """Notify all subscriptions"""
        for sub in self._subs:
            # self.log_it(f"notifying sub {sub.__class__.__name__ }", notify=True)
            sub.post_message(event)

    def _notify_changed(self, changed: SessionChanges) -> None:
        """Notify changed"""
        self._notify_subs(SessionUpdated(session_id=self.id, changed=changed))
        self.post_message(ParSessionUpdated(session_id=self.id, changed=changed))

    @property
    def stats(self) -> TokenStats | None:
        """Get the chat session stats"""
        return self._stream_stats

    @property
    def llm_provider_name(self) -> LlmProvider:
        """Get the LLM model name"""
        return self._llm_config.provider

    @llm_provider_name.setter
    def llm_provider_name(self, value: LlmProvider) -> None:
        """Set the LLM model name"""
        if self._llm_config.provider == value:
            return
        self._llm_config.provider = value
        self._stream_stats = None
        self._changes.add("provider")
        self._changes.add("model")
        self.save()

    @property
    def llm_model_name(self) -> str:
        """Get the LLM model name"""
        return self._llm_config.model_name

    @llm_model_name.setter
    def llm_model_name(self, value: str) -> None:
        """Set the LLM model name"""
        value = value.strip()
        if self._llm_config.model_name == value:
            return
        self._llm_config.model_name = value
        self._stream_stats = None
        self._changes.add("model")
        self.save()

    @property
    def temperature(self) -> float:
        """Get the temperature"""
        return self._llm_config.temperature

    @temperature.setter
    def temperature(self, value: float) -> None:
        """Set the temperature"""
        self._llm_config.temperature = value
        self._changes.add("temperature")
        self.save()

    @property
    def num_ctx(self) -> int | None:
        """Get the number of context tokens"""
        return self._llm_config.num_ctx

    @num_ctx.setter
    def num_ctx(self, value: int | None) -> None:
        """Set the number of context tokens"""
        self._llm_config.num_ctx = value
        self._changes.add("num_ctx")
        self.save()

    # pylint: disable=too-many-branches, too-many-statements
    async def send_chat(self, from_user: str) -> bool:
        """Send a chat message to LLM"""
        self._generating = True
        is_aborted = False
        msg: ParllamaChatMessage | None = None
        try:
            if from_user:
                # self.log_it("CM adding user message")
                msg = ParllamaChatMessage(role="user", content=from_user)
                self.add_message(msg)
                self._notify_subs(ChatMessage(parent_id=self.id, message_id=msg.id, is_final=True))
                self.post_message(ParChatUpdated(parent_id=self.id, message_id=msg.id, is_final=True))
                self.save()

            num_tokens: int = 0
            start_time = datetime.now(UTC)
            ttft: float = 0.0  # time to first token

            # self.log_it(self._llm_config)
            chat_history = [m.to_langchain_native() for m in self.messages]
            # self.log_it(chat_history)
            chat_model = self._llm_config.build_chat_model()
            stream: Iterator[BaseMessageChunk] = chat_model.stream(
                chat_history,  # type: ignore
                config=llm_run_manager.get_runnable_config(chat_model.name or ""),
            )
            # self.log_it("CM adding assistant message")
            msg = ParllamaChatMessage(role="assistant")
            self.add_message(msg)
            self._notify_subs(ChatMessage(parent_id=self.id, message_id=msg.id))
            self.post_message(ParChatUpdated(parent_id=self.id, message_id=msg.id))
            try:
                for chunk in stream:
                    # self.log_it(chunk)
                    elapsed_time = datetime.now(UTC) - start_time
                    if chunk.content:
                        if num_tokens == 0:
                            ttft = elapsed_time.total_seconds()
                        num_tokens += 1
                        if isinstance(chunk.content, str):
                            msg.content += chunk.content

                    if self._abort:
                        is_aborted = True
                        try:
                            msg.content += "\n\nAborted..."
                            self._notify_subs(ChatGenerationAborted(self.id))
                            stream.close()  # type: ignore
                        except Exception:  # pylint:disable=broad-except
                            pass
                        finally:
                            self._abort = False
                        break

                    if (
                        hasattr(chunk, "usage_metadata") and chunk.usage_metadata  # pyright: ignore [reportAttributeAccessIssue]
                    ):
                        # self.log_it(chunk)
                        usage_metadata = (
                            chunk.usage_metadata  # pyright: ignore [reportAttributeAccessIssue]
                        )
                        self._stream_stats = TokenStats(
                            model=self._llm_config.model_name,
                            created_at=datetime.now(),
                            total_duration=int(elapsed_time.total_seconds()),
                            load_duration=0,
                            prompt_eval_count=usage_metadata["input_tokens"],
                            prompt_eval_duration=0,
                            eval_count=usage_metadata["output_tokens"],
                            eval_duration=int(elapsed_time.total_seconds() - ttft),
                            input_tokens=usage_metadata["input_tokens"],
                            output_tokens=usage_metadata["output_tokens"],
                            total_tokens=usage_metadata["total_tokens"],
                            time_til_first_token=int(ttft),
                        )
                        # self.log_it(self._stream_stats)
                    if hasattr(chunk, "response_metadata"):
                        if "model" in chunk.response_metadata:
                            self._stream_stats = TokenStats(
                                model=chunk.response_metadata.get("model") or "?",
                                created_at=chunk.response_metadata.get("created_at") or datetime.now(),
                                total_duration=chunk.response_metadata.get("total_duration") or 0,
                                load_duration=chunk.response_metadata.get("load_duration") or 0,
                                prompt_eval_count=chunk.response_metadata.get("prompt_eval_count") or 0,
                                prompt_eval_duration=chunk.response_metadata.get("prompt_eval_duration") or 0,
                                eval_count=chunk.response_metadata.get("eval_count") or 0,
                                eval_duration=int(chunk.response_metadata.get("eval_duration", 0) / 1_000_000_000) or 0,
                                input_tokens=0,
                                output_tokens=0,
                                total_tokens=0,
                                time_til_first_token=int(ttft),
                            )
                    self._notify_subs(ChatMessage(parent_id=self.id, message_id=msg.id, is_final=not chunk.content))
                    self.post_message(ParChatUpdated(parent_id=self.id, message_id=msg.id, is_final=not chunk.content))
            except Exception as e:  # pylint: disable=broad-except
                err_msg = str(e)
                if self._llm_config.provider == LlmProvider.LLAMACPP and err_msg.startswith(
                    "object of type 'NoneType' has no len()"
                ):
                    self._notify_subs(ChatMessage(parent_id=self.id, message_id=msg.id, is_final=True))
                    self.post_message(ParChatUpdated(parent_id=self.id, message_id=msg.id, is_final=True))
                else:
                    self.log_it(e)
                    self.log_it("Error generating message", notify=True, severity="error")
                    if msg is not None:
                        # err_msg = f"{err_msg}\n{traceback.format_exc()}"
                        if err_msg[18:].startswith("{"):
                            err_dict = ast.literal_eval(err_msg[18:])
                            err_msg = err_dict.get("error", {}).get("message") or err_msg

                        msg.content += f"\n\n{err_msg}"
                        msg.content = msg.content.strip()

            self._changes.add("messages")
            self.save()

            if (
                not is_aborted
                and settings.auto_name_session
                and settings.auto_name_session_llm_config
                and not self.name_generated
            ):
                self.name_generated = True
                user_msg = self.get_first_user_message()
                if user_msg and user_msg.content:
                    self.log_it("Auto naming session", notify=True)
                    self.post_message(
                        ParSessionAutoName(
                            session_id=self.id,
                            llm_config=LlmConfig.from_json(settings.auto_name_session_llm_config),
                            context=user_msg.content,
                        )
                    )
        except Exception as e:  # pylint: disable=broad-except
            self.log_it(e)
            self.log_it("Error generating message", notify=True, severity="error")
            if msg is not None:
                err_msg = str(e)
                if err_msg[18:].startswith("{"):
                    err_dict = ast.literal_eval(err_msg[18:])
                    err_msg = err_dict.get("error", {}).get("message") or err_msg

                # err_msg = f"{err_msg}\n{traceback.format_exc()}"

                msg.content += f"\n\n{err_msg}"
                msg.content = msg.content.strip()
                self._changes.add("messages")
                self.save()
                self._notify_subs(ChatMessage(parent_id=self.id, message_id=msg.id, is_final=True))
                return False
        finally:
            self._generating = False

        return not is_aborted

    def new_session(self, name: str = "My Chat"):
        """Start new session"""
        self._batching = True
        self.id = uuid.uuid4().hex
        self.name = name
        self.name_generated = False
        self.messages.clear()
        self._id_to_msg.clear()
        self.clear_changes()
        self._loaded = False
        self._stream_stats = None
        self._batching = False

    def __eq__(self, other: object) -> bool:
        """Check if two sessions are equal"""
        if not isinstance(other, ChatSession):
            return NotImplemented
        return self.id == other.id

    def __ne__(self, other: object) -> bool:
        """Check if two sessions are not equal"""
        if not isinstance(other, ChatSession):
            return NotImplemented
        return self.id != other.id

    def to_json(self) -> str:
        """Convert the chat session to JSON"""
        return json.dumps(
            {
                "id": self.id,
                "_salt": (base64.b64encode(self._salt).decode("utf-8") if self._salt else None),
                "__key__": self._key_secure,
                "name": self.name,
                "name_generated": self.name_generated,
                "last_updated": self.last_updated.isoformat(),
                "llm_config": self._llm_config.to_json(),
                "messages": [m.to_dict() for m in self.messages],
            },
            str,
            json.OPT_INDENT_2,
        ).decode("utf-8")

    @staticmethod
    def from_json(json_data: str, load_messages: bool = False) -> ChatSession:
        """Convert JSON to chat session"""
        data: dict = json.loads(json_data)
        if load_messages:
            messages = data["messages"]
            for m in messages:
                # convert old format
                if "message_id" in m:
                    m["id"] = "message_id"
                    del m["message_id"]
        else:
            messages = []
        utc = pytz.UTC
        lc = data.get("llm_config")
        # adapt old format session
        if not lc:
            lc = LlmConfig(
                provider=LlmProvider.OLLAMA,
                model_name=data["llm_model_name"],
                temperature=data.get("options", {}).get("temperature", 0.5),
            ).to_json()
        session = ChatSession(
            id=data.get("id", data.get("id", data.get("session_id"))),
            name=data.get("name", data.get("name", data.get("session_name"))),
            last_updated=datetime.fromisoformat(data["last_updated"]).replace(tzinfo=utc),
            messages=([ParllamaChatMessage(**m) for m in messages] if load_messages else None),
            llm_config=LlmConfig.from_json(lc),
        )
        session.name_generated = True

        if salt := data.get("__salt__"):
            session._salt = base64.b64decode(salt)  # pylint: disable=protected-access
        else:
            session._salt = None  # pylint: disable=protected-access

        session._key_secure = data.get("__key__")  # pylint: disable=protected-access

        return session

    @staticmethod
    def load_from_file(filename: str) -> ChatSession | None:
        """Load a chat session from a file"""
        try:
            with open(os.path.join(settings.chat_dir, filename), encoding="utf-8") as f:
                return ChatSession.from_json(f.read())
        except OSError:
            return None

    @property
    def is_valid(self) -> bool:
        """return true if session is valid"""
        return (
            len(self.name) > 0 and len(self.llm_model_name) > 0 and self.llm_model_name not in ["Select.BLANK", "None"]
            # and len(self.messages) > 0
        )

    def save(self) -> bool:
        """Save the chat session to a file"""
        if self._batching:
            # self.log_it(f"CS is batching, not notifying: {self.name}")
            return False
        if not self._loaded:
            self.load()
        if not self.is_dirty:
            # self.log_it(f"CS is not dirty, not notifying: {self.name}")
            return False  # No need to save if no changes

        self.last_updated = datetime.now(UTC)
        if "system_prompt" in self._changes:
            msg = self.system_prompt
            if msg is not None:
                self._notify_subs(ChatMessage(parent_id=self.id, message_id=msg.id))
        nc: SessionChanges = SessionChanges()
        for change in self._changes:
            if change in session_change_list:
                nc.add(change)  # type: ignore

        self._notify_changed(nc)
        self.clear_changes()

        if settings.no_save_chat:
            return False  # Do not save if no_save_chat is set in settings
        if not self.is_valid or len(self.messages) == 0:
            # self.log_it(f"CS not valid, not saving: {self.id}")
            return False  # Cannot save without name, LLM model name and at least one message

        # self.log_it(f"CS saving: {self.name}")

        file_name = f"{self.id}.json"  # Use session ID as filename to avoid over
        try:
            with open(os.path.join(settings.chat_dir, file_name), "w", encoding="utf-8") as f:
                f.write(self.to_json())
            return True
        except OSError:
            return False

    def stop_generation(self) -> None:
        """Stop LLM model generation"""
        self._abort = True

    @property
    def abort_pending(self) -> bool:
        """Check if LLM model generation is pending"""
        return self._abort

    @property
    def is_generating(self) -> bool:
        """Check if LLM model generation is in progress"""
        return self._generating

    @property
    def num_subs(self):
        """Return the number of subscribers"""
        return len(self._subs)

    def on_par_chat_message_deleted(self, event: ParChatMessageDeleted):
        """Handle ParChatMessageDeleted event"""
        self._notify_subs(ChatMessageDeleted(parent_id=event.parent_id, message_id=event.message_id))



"""Shared types"""

from __future__ import annotations

from typing import Literal, TypeAlias

# TODO change to enums
SessionChanges: TypeAlias = set[Literal["name", "provider", "model", "temperature", "options", "messages", "num_ctx"]]
session_change_list: list[Literal["name", "provider", "model", "temperature", "options", "messages", "num_ctx"]] = [
    "name",
    "provider",
    "model",
    "temperature",
    "options",
    "messages",
    "num_ctx",
]

PromptChanges: TypeAlias = set[Literal["name", "description", "messages", "submit_on_load"]]
prompt_change_list: list[Literal["name", "description", "messages", "submit_on_load"]] = [
    "name",
    "description",
    "messages",
    "submit_on_load",
]



"""Provides a modal dialog for prompting for a save screen."""

from __future__ import annotations

from pathlib import Path
from typing import Any

from textual.app import App
from textual_fspicker import FileSave, Filters


##############################################################################
class SaveSession(FileSave):
    """Modal dialog for prompting for the name of a conversation file."""

    def __init__(self) -> None:
        super().__init__(
            ".",
            filters=Filters(
                (
                    "Markdown",
                    lambda p: p.suffix.lower() in (".md", ".markdown"),
                ),
                ("Text", lambda p: p.suffix.lower() in (".txt", ".text")),
                ("Any", lambda _: True),
            ),
        )

    @classmethod
    async def get_filename(cls, app: App[Any]) -> Path | None:
        """Get the filename from the user.

        Args:
            app: The app.

        Returns:

            The path of the file the user selected, or `None` if they
            cancelled.
        """
        return await app.push_screen_wait(cls())



"""Check for PARLLAMA Updates on pypi"""

from __future__ import annotations

from datetime import UTC, datetime

import httpx
import requests
from semver import Version

from parllama import __version__
from parllama.par_event_system import ParEventSystemBase
from parllama.settings_manager import settings


class UpdateManager(ParEventSystemBase):
    """Update manager class"""

    url: str

    def __init__(self) -> None:
        """Initialize the update manager"""
        super().__init__(id="update_manager")
        self.url = "https://pypi.org/pypi/parllama/json"

    async def get_latest_version(self) -> Version:
        """Get project version"""
        self.log_it("Checking for updates...")

        try:
            async with httpx.AsyncClient() as client:
                response = await client.get(self.url, follow_redirects=True, timeout=5)
                data = response.json()
            return Version.parse(data["info"]["version"])
        except requests.exceptions.RequestException as e:
            raise ValueError("Could not fetch version info") from e

    async def check_for_updates(self, force: bool = False) -> None:
        """Check for updates"""
        self.log_it(f"App version is: {__version__}")
        if not force:
            last_version_check = settings.last_version_check
            if last_version_check:
                if (datetime.now(UTC) - last_version_check).days < 1:
                    return
        try:
            latest_version = await self.get_latest_version()
            self.log_it(f"Latest version is: {latest_version}")

            if Version.parse(__version__) < latest_version:
                if self.app:
                    self.log_it(
                        f"New version available: {latest_version}",
                        notify=True,
                        timeout=8,
                    )
                else:
                    print(f"New version available: {latest_version}")
            settings.last_version_check = datetime.now(UTC)
            settings.save()
        except (ValueError, TypeError) as e:
            if self.app:
                self.log_it(e, notify=True, severity="error")
            else:
                print(e)


update_manager = UpdateManager()



"""Widget for chatting with LLM."""

from __future__ import annotations

import re
from pathlib import Path
from typing import cast

from par_ai_core.llm_providers import (
    LlmProvider,
    get_provider_name_fuzzy,
    llm_provider_names,
)
from textual import on
from textual.app import ComposeResult
from textual.binding import Binding
from textual.containers import Horizontal, Vertical
from textual.events import Show
from textual.suggester import SuggestFromList
from textual.widgets import Button, ContentSwitcher, Select, TabbedContent

from parllama.chat_manager import ChatSession, chat_manager
from parllama.chat_message import ParllamaChatMessage
from parllama.dialogs.information import InformationDialog
from parllama.messages.messages import (
    ChangeTab,
    ChatGenerationAborted,
    ChatMessage,
    ChatMessageSent,
    ClearChatInputHistory,
    DeleteSession,
    LogIt,
    PromptListChanged,
    PromptListLoaded,
    PromptSelected,
    ProviderModelsChanged,
    RegisterForUpdates,
    SessionSelected,
    SessionToPrompt,
    SessionUpdated,
    UpdateChatControlStates,
    UpdateTabLabel,
)
from parllama.provider_manager import provider_manager
from parllama.settings_manager import settings
from parllama.widgets.session_config import SessionConfig
from parllama.widgets.session_list import SessionList
from parllama.widgets.user_input import UserInput
from parllama.widgets.views.chat_tab import ChatTab

valid_commands: list[str] = [
    "/help",
    "/?",
    "/tab.",
    "/tab.remove",
    "/tabs.clear",
    "/session.",
    "/session.provider",
    "/session.model",
    "/session.temp",
    "/session.new",
    "/session.name",
    "/session.delete",
    "/session.export",
    "/session.system_prompt",
    "/session.clear_system_prompt",
    "/session.to_prompt",
    "/history.clear",
    "/add.image",
    "/prompt.load ",
]


class ChatView(Vertical, can_focus=False, can_focus_children=True):
    """Widget for viewing application logs."""

    DEFAULT_CSS = """
    ChatView {
        layers: left;
        SessionList {
            width: 40;
            height: 1fr;
            dock: left;
            padding: 1;
        }
        #chat_tabs {
            height: 1fr;
        }
        #send_bar {
            height: auto;
            min-height: 3;
            max-height: 15;
            background: $surface-darken-1;
            #send_button {
                min-width: 7;
                width: 7;
                margin-right: 1;
            }
            #stop_button {
                min-width: 6;
                width: 6;
            }
        }
    }
    """

    BINDINGS = [
        Binding(
            key="ctrl+s",
            action="toggle_session_list",
            description="Sessions",
            show=True,
            priority=True,
        ),
        Binding(
            key="ctrl+b",
            action="new_session",
            description="New Session",
            show=True,
            priority=True,
        ),
        Binding(
            key="ctrl+n",
            action="new_tab",
            description="New Tab",
            show=True,
        ),
        Binding(
            key="ctrl+delete",
            action="remove_tab",
            description="Remove Tab",
            key_display="^del",
            show=True,
        ),
        Binding(
            key="ctrl+e",
            action="export",
            description="Export",
            show=True,
            priority=True,
        ),
        Binding(
            key="ctrl+p",
            action="toggle_session_config",
            description="Config",
            show=True,
            priority=True,
        ),
    ]
    chat_tabs: TabbedContent
    provider_list_auto_complete_list: list[str]
    model_list_auto_complete_list: list[str]
    prompt_list_auto_complete_list: list[str]

    def __init__(self, **kwargs) -> None:
        """Initialise the view."""
        super().__init__(**kwargs)

        self.session_list = SessionList()
        self.session_list.display = False

        self.chat_tabs = TabbedContent(id="chat_tabs")
        self.user_input: UserInput = UserInput(
            id="user_input",
            suggester=SuggestFromList(
                valid_commands,
                case_sensitive=False,
            ),
            history_file=Path(settings.chat_history_file),
        )

        self.send_button: Button = Button("Send", id="send_button", disabled=True, variant="success")
        self.stop_button: Button = Button("Stop", id="stop_button", disabled=True, variant="error")
        self.last_command = ""
        self.provider_list_auto_complete_list = [f"/session.provider {p}" for p in llm_provider_names]
        self.model_list_auto_complete_list = []
        self.prompt_list_auto_complete_list = []

    def compose(self) -> ComposeResult:
        """Compose the content of the view."""

        yield self.session_list
        with self.chat_tabs:
            yield ChatTab(user_input=self.user_input, session_list=self.session_list)
        with Horizontal(id="send_bar"):
            yield self.user_input
            yield self.send_button
            yield self.stop_button

    async def on_mount(self) -> None:
        """Set up the dialog once the DOM is ready."""
        self.app.post_message(
            RegisterForUpdates(
                widget=self,
                event_names=[
                    "ProviderModelsChanged",
                    "DeleteSession",
                    "SessionSelected",
                    "PromptListLoaded",
                    "PromptListChanged",
                    "PromptSelected",
                ],
            )
        )
        self.prompt_list_auto_complete_list = [f"/prompt.load {prompt.name}" for prompt in chat_manager.sorted_prompts]

    def _on_show(self, event: Show) -> None:
        """Handle show event"""
        self.user_input.focus()

    @on(PromptListLoaded)
    def on_prompt_list_loaded(self, event: PromptListLoaded) -> None:
        """Prompt list changed"""
        event.stop()
        self.post_message(LogIt("Prompt list loaded"))
        self.prompt_list_auto_complete_list = [f"/prompt.load {prompt.name}" for prompt in chat_manager.sorted_prompts]
        self.rebuild_suggester()

    @on(UserInput.Changed)
    def on_user_input_changed(self, event: UserInput.Changed) -> None:
        """Handle max lines input change"""
        event.stop()
        self.update_control_states()

    def update_control_states(self):
        """Update disabled state of controls based on model and user input values"""
        self.send_button.disabled = (
            self.active_tab.busy
            or self.active_tab.session_config.provider_model_select.provider_select.value == Select.BLANK
            or self.active_tab.session_config.provider_model_select.model_select.value == Select.BLANK
            or len(self.user_input.value.strip()) == 0
        )
        self.stop_button.disabled = not self.active_tab.busy or self.session.abort_pending

    @on(ProviderModelsChanged)
    def model_list_changed(self, evt: ProviderModelsChanged) -> None:
        """Model list changed"""
        evt.stop()
        if not evt.provider or self.session_config.provider_model_select.provider_select.value != evt.provider:
            return
        self.post_message(LogIt("Provider models changed"))
        self.model_list_auto_complete_list = [
            f"/session.model {m}"
            for m in provider_manager.get_model_names(
                self.active_tab.session_config.provider_model_select.provider_select.value  # type: ignore
            )
        ]
        self.rebuild_suggester()
        self.update_control_states()

    @on(PromptListChanged)
    def on_prompt_list_changed(self, evt: PromptListChanged) -> None:
        """Prompt list changed"""
        evt.stop()
        self.post_message(LogIt("Prompt list changed"))
        self.prompt_list_auto_complete_list = [f"/prompt.load {prompt.name}" for prompt in chat_manager.sorted_prompts]
        self.rebuild_suggester()

    def rebuild_suggester(self) -> None:
        """Rebuild the suggester"""
        self.user_input.suggester = SuggestFromList(
            valid_commands
            + self.provider_list_auto_complete_list
            + self.model_list_auto_complete_list
            + self.prompt_list_auto_complete_list,
            case_sensitive=False,
        )

    @on(Button.Pressed, "#new_button")
    async def on_new_button_pressed(self, event: Button.Pressed) -> None:
        """New button pressed"""
        event.stop()
        await self.action_new_tab()

    @on(Button.Pressed, "#send_button")
    async def on_send_button_pressed(self, event: Button.Pressed) -> None:
        """Send button pressed"""
        event.stop()
        self.user_input.submit()

    @on(UserInput.Submitted)
    async def action_send_message(self, event: UserInput.Submitted) -> None:
        """Send the message."""
        event.stop()
        if settings.close_session_config_on_submit:
            self.active_tab.session_config.display = False

        self.user_input.focus()

        if self.send_button.disabled:
            return

        user_msg: str = event.value.strip()
        if not user_msg:
            return

        if self.active_tab.busy:
            self.notify("LLM Busy...", severity="error")
            return
        self.last_command = user_msg
        self.user_input.value = ""
        self.update_control_states()

        if user_msg.startswith("/"):
            return await self.handle_command(user_msg[1:].strip())

        self.active_tab.do_send_message(user_msg)

    # pylint: disable=too-many-statements,too-many-branches,too-many-return-statements
    async def handle_command(self, cmd_raw: str) -> None:
        """Handle a command"""
        cmd = cmd_raw.lower()
        if cmd in ("?", "help"):
            await self.app.push_screen(
                InformationDialog(
                    title="Chat Commands",
                    message="""
Chat Commands:
/tab.# - Switch to the tab with the given number
/tab.new - Create new tab and switch to it
/tab.remove - Remove the active tab
/tabs.clear - Clear / remove all tabs
/session.new [session_name] - Start new chat session in current tab with optional name
/session.name [session_name] - Select session name input or set the session name in current tab
/session.provider [provider_name] - Select the AI provider dropdown or set provider name in current tab
/session.model [model_name] - Select model dropdown or set model name in current tab
/session.temp [temperature] - Select temperature input or set temperature in current tab
/session.delete - Delete the chat session for current tab
/session.export - Export the conversation in current tab to a Markdown file
/session.system_prompt [system_prompt] - Set system prompt in current tab
/session.clear_system_prompt - Remove system prompt in current tab
/session.to_prompt submit_on_load [prompt_name] - Copy current session to new custom prompt. submit_on_load = {0|1}
/prompt.load prompt_name - Load a custom prompt using current tabs model and temperature
/add.image image_path_or_url prompt - Add an image via path or url to the active chat session
/history.clear - Clear chat input history
                        """,
                )
            )
            return

        tab_number_re = re.compile(r"^t(?:ab)?\.(\d+)$")
        if match := tab_number_re.match(cmd):
            (tab_number,) = match.groups()
            tab_number = int(tab_number) - 1
            if 0 <= tab_number <= len(self.chat_tabs.children):
                self.focus_tab(tab_number)
            else:
                self.notify(f"Tab {tab_number + 1} does not exist", severity="error")
            return
        if cmd == "tab.new":
            await self.action_new_tab()
        elif cmd == "tab.remove":
            await self.remove_active_tab()
        elif cmd == "tabs.clear":
            await self.remove_all_tabs()
        elif cmd == "session.new":
            await self.active_tab.action_new_session()
        elif cmd.startswith("session.new "):
            (_, v) = cmd.split(" ", 1)
            v = v.strip()
            await self.active_tab.action_new_session(v)
        elif cmd == "session.delete":
            self.app.post_message(DeleteSession(session_id=self.session.id))
        elif cmd == "session.provider":
            self.active_tab.session_config.display = True
            self.set_timer(
                0.1,
                self.active_tab.session_config.provider_model_select.provider_select.focus,
            )
            return
        elif cmd.startswith("session.provider "):
            (_, v) = cmd.split(" ", 1)
            v_org = v.strip()
            v = get_provider_name_fuzzy(v_org)
            if not v:
                self.notify(
                    f"Provider {v_org} not found in [{','.join(llm_provider_names)}]",
                    severity="error",
                )
                return
            self.active_tab.session_config.provider_model_select.provider_select.value = LlmProvider(v)
            self.set_timer(0.1, self.user_input.focus)
        elif cmd == "session.model":
            self.active_tab.session_config.display = True
            self.set_timer(
                0.1,
                self.active_tab.session_config.provider_model_select.model_select.focus,
            )
            return
        elif cmd.startswith("session.model "):
            (_, v) = cmd_raw.split(" ", 1)
            v = provider_manager.get_model_name_fuzzy(
                self.active_tab.session_config.provider_model_select.provider_select.value,  # type: ignore
                v.strip(),
            )
            if not v:
                self.notify(f"Model {v} not found", severity="error")
                return
            self.active_tab.session_config.provider_model_select.model_select.value = v
            self.set_timer(0.1, self.user_input.focus)
        elif cmd == "session.temp":
            self.set_timer(0.1, self.active_tab.session_config.temperature_input.focus)
            self.active_tab.session_config.display = True
        elif cmd.startswith("session.temp "):
            (_, v) = cmd.split(" ", 1)
            v = v.strip()
            self.active_tab.session_config.temperature_input.value = v
            self.set_timer(0.1, self.user_input.focus)
        elif cmd == "session.name":
            self.set_timer(0.1, self.active_tab.session_config.session_name_input.focus)
            self.active_tab.session_config.display = True
        elif cmd.startswith("session.name "):
            (_, v) = cmd.split(" ", 1)
            v = v.strip()
            self.active_tab.session_config.session_name_input.value = v
            await self.active_tab.session_config.session_name_input.action_submit()
        elif cmd.startswith("session.to_prompt "):
            vs: list[str] = cmd.split(" ", 2)
            if len(vs) == 2:
                (_, submit_on_load) = vs
                v = ""
            else:
                (_, submit_on_load, v) = vs
            v = v.strip()
            self.app.post_message(
                SessionToPrompt(
                    session_id=self.session.id,
                    submit_on_load=submit_on_load == "1",
                    prompt_name=v,
                )
            )
        elif cmd.startswith("session.export"):
            self.active_tab.save_conversation_text()
        elif cmd.startswith("session.clear_system_prompt"):
            self.session.system_prompt = None
        elif cmd.startswith("session.system_prompt "):
            (_, v) = cmd.split(" ", 1)
            v = v.strip()
            if not v:
                self.notify("System prompt cannot be empty", severity="error")
                return
            self.session.system_prompt = ParllamaChatMessage(role="system", content=v)
            self.active_tab.post_message(
                ChatMessage(
                    parent_id=self.session.id,
                    message_id=self.session.system_prompt.id,  # pyright: ignore [reportOptionalMemberAccess]
                    is_final=True,
                )
            )
        elif cmd.startswith("prompt.load "):
            (_, v) = cmd.split(" ", 1)
            v = v.strip()
            prompt = chat_manager.get_prompt_by_name(v)
            if prompt is None:
                self.notify(f"Prompt {v} not found", severity="error")
                return
            await self.active_tab.load_prompt(
                PromptSelected(
                    prompt_id=prompt.id,
                    model_name=None,
                    temperature=None,
                )
            )
        elif cmd.startswith("add.image "):
            parts = cmd.split(" ")
            if len(parts) < 3:
                self.notify("Usage add.image FILE_NAME PROMPT", severity="error")
            v = parts[1].strip()
            p = " ".join(parts[2:])
            if v.startswith("http://") or v.startswith("https://"):
                msg = ParllamaChatMessage(role="user", content=p, images=[v])
            else:
                path = Path(v)
                if not path.exists():
                    self.notify(f"Image {v} not found", severity="error")
                    return
                msg = ParllamaChatMessage(role="user", content=p, images=[str(path.absolute())])

            self.session.add_message(msg)
            self.post_message(ChatMessage(parent_id=self.session.id, message_id=msg.id))
            self.active_tab.do_send_message("")
        elif cmd == "history.clear":
            self.app.post_message(ClearChatInputHistory())
        else:
            self.notify(f"Unknown command: {cmd}", severity="error")

    @on(Button.Pressed, "#stop_button")
    def on_stop_button_pressed(self, event: Button.Pressed) -> None:
        """Stop button pressed"""
        event.stop()
        self.stop_button.disabled = True
        self.active_tab.session.stop_generation()
        # self.workers.cancel_group(self, "message_send")
        self.workers.cancel_node(self.active_tab)
        self.active_tab.busy = False

    @on(ChatMessageSent)
    def on_chat_message_sent(self, event: ChatMessageSent) -> None:
        """Handle a chat message sent"""
        event.stop()
        if self.session.id == event.session_id:
            self.stop_button.disabled = True

    @on(SessionUpdated)
    async def session_updated(self, event: SessionUpdated) -> None:
        """Session updated event"""
        event.stop()
        session = chat_manager.get_session(event.session_id)
        if session != self.session:
            return
        # self.notify(f"Chat View session updated {','.join([*event.changed])}")
        if "provider" in event.changed:
            self.model_list_changed(ProviderModelsChanged())

    def action_toggle_session_list(self) -> None:
        """Toggle the session list."""
        self.session_list.display = not self.session_list.display
        if self.session_list.display:
            self.set_timer(0.1, self.session_list.list_view.focus)
        else:
            self.set_timer(0.1, self.user_input.focus)

    @on(SessionSelected)
    async def session_selected(self, event: SessionSelected) -> None:
        """Session selected event"""
        event.stop()
        if event.new_tab:
            await self.action_new_tab()
            await self.active_tab.load_session(event.session_id)
        else:
            await self.active_tab.load_session(event.session_id)

    @on(PromptSelected)
    async def prompt_selected(self, event: PromptSelected) -> None:
        """Prompt selected event"""
        event.stop()
        await self.action_new_tab()
        await self.active_tab.load_prompt(event)
        self.app.post_message(ChangeTab(tab="Chat"))

    @on(UpdateChatControlStates)
    def update_chat_control_states(self, event: UpdateChatControlStates) -> None:
        """Update chat control states event"""
        event.stop()
        self.update_control_states()

    @property
    def active_tab(self) -> ChatTab:
        """Return the active chat tab."""
        return cast(ChatTab, self.chat_tabs.get_pane(self.chat_tabs.active))

    @property
    def session(self) -> ChatSession:
        """Return the active chat session."""
        return self.active_tab.session

    @property
    def ai_provider(self) -> LlmProvider:
        """Get the AI provider"""
        v = self.session_config.provider_model_select.provider_select.value
        if v == Select.BLANK:
            return LlmProvider.OLLAMA
        return v  # type: ignore

    @property
    def ai_model(self) -> str:
        """Get the AI model"""
        v = self.session_config.provider_model_select.model_select.value
        if v == Select.BLANK:
            return ""
        return str(v)

    @property
    def session_config(self) -> SessionConfig:
        """Get the session config"""
        return self.active_tab.session_config

    @on(ChatMessage)
    async def on_chat_message(self, event: ChatMessage) -> None:
        """Route chat message to correct tab"""
        event.stop()
        for tab in self.chat_tabs.query(ChatTab):
            if tab.session.id == event.parent_id:
                await tab.on_chat_message(event)

    @on(UpdateTabLabel)
    def on_update_tab_label(self, event: UpdateTabLabel) -> None:
        """Update tab label event"""
        event.stop()
        # self.notify(f"Updated tab label: {event.tab_label}")
        tab = self.chat_tabs.get_tab(event.tab_id)
        tab_num = self.chat_tabs.get_child_by_type(ContentSwitcher).children.index(
            self.chat_tabs.get_pane(event.tab_id)
        )
        tab.label = f"[{tab_num + 1}] " + event.tab_label  # type: ignore

    def re_index_labels(self) -> None:
        """Re-index tab labels"""
        for ctab in self.chat_tabs.query(ChatTab):
            tab = self.chat_tabs.get_tab(str(ctab.id))
            tab_num = self.chat_tabs.get_child_by_type(ContentSwitcher).children.index(
                self.chat_tabs.get_pane(str(ctab.id))
            )
            tab.label = f"[{tab_num + 1}] " + tab.label.plain.split(" ", 1)[1]  # type: ignore

    @on(DeleteSession)
    async def on_delete_session(self, event: DeleteSession) -> None:
        """Delete session event"""
        event.stop()
        self.app.post_message(LogIt(f"Deleted chat session {event.session_id}"))
        tab_removed: bool = False
        for tab in self.chat_tabs.query(ChatTab):
            if tab.session.id == event.session_id:
                await self.chat_tabs.remove_pane(str(tab.id))
                tab_removed = True

        chat_manager.delete_session(event.session_id)
        if len(self.chat_tabs.query(ChatTab)) == 0:
            await self.action_new_tab()
        elif tab_removed:
            self.re_index_labels()

        self.notify("Chat session deleted")
        # self.user_input.focus()

    async def action_new_tab(self) -> None:
        """New tab action"""
        tab = ChatTab(user_input=self.user_input, session_list=self.session_list)
        await self.chat_tabs.add_pane(tab)
        self.chat_tabs.active = str(tab.id)

    async def remove_active_tab(self) -> None:
        """Remove the active tab"""
        await self.chat_tabs.remove_pane(self.chat_tabs.active)
        if len(self.chat_tabs.query(ChatTab)) == 0:
            await self.action_new_tab()
        else:
            self.re_index_labels()

    async def remove_all_tabs(self) -> None:
        """Remove all tabs"""
        for tab in self.chat_tabs.query(ChatTab):
            await self.chat_tabs.remove_pane(str(tab.id))
        await self.action_new_tab()

    def action_export(self) -> None:
        """Export chat to Markdown file"""
        self.active_tab.save_conversation_text()

    async def action_new_session(self) -> None:
        """Start a new chat session"""
        await self.active_tab.action_new_session()

    async def action_remove_tab(self) -> None:
        """Remove the active tab"""
        await self.remove_active_tab()

    def focus_tab(self, tab_num: int) -> None:
        """Focus requested tab"""
        tabs = self.chat_tabs.query(ChatTab)
        if len(tabs) > tab_num:
            self.chat_tabs.active = str(tabs[tab_num].id)

    @on(ChatGenerationAborted)
    def on_chat_generation_aborted(self, event: ChatGenerationAborted) -> None:
        """Chat generation aborted event"""
        event.stop()
        self.notify("Chat Aborted", severity="warning")

    @on(TabbedContent.TabActivated)
    def on_tab_activated(self, msg: TabbedContent.TabActivated) -> None:
        """Prevent Tab activated event bubble"""
        msg.stop()

    def action_toggle_session_config(self) -> None:
        """Toggle session configuration panel"""
        self.active_tab.action_toggle_session_config()



"""Validator for URLs."""

from __future__ import annotations

import urllib.parse

from textual.validation import ValidationResult, Validator


class HttpValidator(Validator):
    """Validator for URLs."""

    def validate(self, value: str) -> ValidationResult:
        """Validate if the input is a valid URL."""
        try:
            value = value.lower()
            if not (value.startswith("http://") or value.startswith("https://")):
                raise ValueError("Must start with http:// or https://")

            result = urllib.parse.urlparse(value)
            if not result.scheme or not result.netloc:
                raise ValueError("Invalid URL")
            if " " in value:
                raise ValueError("Invalid URL")
            if "." not in value and "localhost" not in value:
                raise ValueError("Invalid URL")
        except ValueError:
            return self.failure(description="Invalid URL", value=value)
        return ValidationResult.success()



"""Chat message class"""

from __future__ import annotations

import uuid
from collections.abc import Sequence
from dataclasses import dataclass
from typing import Any, Literal

from ollama import Message as OMessage
from par_ai_core.llm_image_utils import b64_encode_image

from parllama.messages.par_chat_messages import ParChatUpdated
from parllama.models.ollama_data import MessageRoles, ToolCall
from parllama.par_event_system import ParEventSystemBase
from parllama.settings_manager import fetch_and_cache_image


def try_get_image_type(image_path: str) -> Literal["jpeg", "png", "gif"]:
    """Get image type from image path."""
    if image_path.startswith("data:"):
        ext = image_path.split(";")[0].split("/")[-1].lower()
    else:
        ext = image_path.split(".")[-1].lower()
    if ext in ["jpg", "jpeg"]:
        return "jpeg"
    if ext in ["png"]:
        return "png"
    if ext in ["gif"]:
        return "gif"
    raise ValueError(f"Unsupported image type: {ext}")


def image_to_base64(image_bytes: bytes, image_type: Literal["jpeg", "png", "gif"] = "jpeg") -> str:
    """Convert an image to base64 url."""
    return f"data:image/{image_type};base64,{b64_encode_image(image_bytes)}"


def image_to_chat_message(image_url_str: str) -> dict[str, Any]:
    """Convert an image to a chat message."""
    return {
        "type": "image_url",
        "image_url": {"url": image_url_str},
    }


@dataclass
class ParllamaChatMessage(ParEventSystemBase):
    """Chat message."""

    role: MessageRoles
    "Assumed role of the message. Response messages always has role 'assistant'."

    content: str = ""
    "Content of the message. Response messages contains message fragments when streaming."

    images: list[str] | None = None
    """
      Optional list of image data for multimodal models.

      Valid input types are:

      - `str` or path-like object: path to image file
      - `bytes` or bytes-like object: raw image data

      Valid image formats depend on the model. See the model card for more information.
      """

    tool_calls: Sequence[ToolCall] | None = None
    """
    Tools calls to be made by the model.
    """

    def __init__(  # pylint: disable=too-many-arguments
        self,
        *,
        id: str | None = None,  # pylint: disable=redefined-builtin
        role: MessageRoles,
        content: str = "",
        images: list[str] | None = None,
        tool_calls: Sequence[ToolCall] | None = None,
    ) -> None:
        """Initialize the chat message"""
        super().__init__(id=id)
        self.role = role
        self.content = content
        self.images = images
        self.tool_calls = tool_calls

        if self.images:
            image = self.images[0]
            if not image.startswith("data:"):
                image_type = try_get_image_type(image)
                image = image_to_base64(fetch_and_cache_image(image)[1], image_type)
                self.images[0] = image

    def __str__(self) -> str:
        """Ollama message representation"""
        return f"## {self.role}\n\n{self.content}\n\n"

    def to_dict(
        self,
    ) -> dict[str, Any]:
        """Convert the chat message to a dictionary"""
        return {
            "id": self.id,
            "role": self.role,
            "content": self.content,
            "images": self.images,
            "tool_calls": self.tool_calls,
        }

    def to_ollama_native(self) -> OMessage:
        """Convert a message to Ollama native format"""
        return OMessage(role=self.role, content=self.content)

    def to_langchain_native(self) -> tuple[str, str | list[dict[str, Any]]]:
        """Convert a message to Langchain native format"""
        content = self.content
        if self.images:
            image = self.images[0]
            try:
                content = [
                    {"type": "text", "text": self.content},
                    image_to_chat_message(image),
                ]
            except Exception as e:  # pylint: disable=broad-except
                content = str(e)
                # content = "Image is missing, ignore this message."
        return (
            self.role,
            content,
        )

    def notify_changes(self) -> None:
        """Notify changes to the chat message"""
        if self.parent is None:
            return
        self.post_message(ParChatUpdated(parent_id=self.parent.id, message_id=self.id))

    def clone(self, new_id: bool = False) -> ParllamaChatMessage:
        """Clone the chat message"""

        return ParllamaChatMessage(
            id=uuid.uuid4().hex if new_id else self.id,
            role=self.role,
            content=self.content,
            images=[*self.images] if self.images else None,
            tool_calls=self.tool_calls,
        )



"""Deferred select widget."""

from collections.abc import Iterable
from typing import Generic

import rich.repr
from rich.console import RenderableType
from textual import events
from textual.message import Message
from textual.widgets import Select
from textual.widgets._select import (
    BLANK,
    NoSelection,
    SelectType,
)  # pylint: disable=unused-import


class DeferredSelect(Generic[SelectType], Select[SelectType]):
    """Deferred select widget."""

    _deferred_value: SelectType | NoSelection = BLANK

    @rich.repr.auto
    class BadDeferredValue(Message):
        """Posted when the select options do not contain the deferred value.

        This message can be handled using a `on_bad_deferred_value` method.
        """

        def __init__(self, select: Select[SelectType], deferred_value: SelectType | NoSelection) -> None:
            """
            Initialize the BadDeferredValue message.
            """
            super().__init__()
            self.select = select
            self.deferred_value = deferred_value

        def __rich_repr__(self) -> rich.repr.Result:
            """repr"""
            yield self.select
            yield self.deferred_value

        @property
        def control(self) -> Select[SelectType]:
            """The Select that sent the message."""
            return self.select

    def __init__(self, *args, **kwargs):
        """Initialise the widget."""
        if "options" in kwargs:
            opts = [o[1] for o in kwargs["options"]]
        else:
            opts = []
        if "value" in kwargs and kwargs["value"] not in opts:
            self._deferred_value = kwargs["value"]
            del kwargs["value"]
        super().__init__(*args, **kwargs)

    def _on_mount(self, _: events.Mount) -> None:
        """Handle mount event."""
        # if self._deferred_value and self._deferred_value != BLANK:
        #     self.post_message(LogIt(f"Deferred value {self._deferred_value}."))

    @property
    def deferred_value(self) -> SelectType | NoSelection:
        """Get the deferred value."""
        return self._deferred_value

    @deferred_value.setter
    def deferred_value(self, value: SelectType | NoSelection) -> None:
        """Set the deferred value."""
        self._deferred_value = value
        if not self._deferred_value or self._deferred_value == BLANK:
            self.value = BLANK
            return
        for opts in self._options:
            if value == opts[1]:
                self.value = self._deferred_value
                self._deferred_value = BLANK
                return
        self.set_timer(0.1, self.set_options)

    def set_options(self, options: Iterable[tuple[RenderableType, SelectType]] | None = None) -> None:
        """Set the options for the Select."""
        old_value = self.value

        if options is not None:
            super().set_options(options)

        opts = [o[1] for o in self._options]
        if len(opts) == 0:
            return

        if old_value != BLANK and old_value in opts:
            with self.prevent(Select.Changed):
                self.value = old_value
                self._deferred_value = BLANK
                return

        if self._deferred_value is BLANK or not self._deferred_value:
            return
        if self._deferred_value in opts:
            with self.prevent(Select.Changed):
                self.value = self._deferred_value
            self._deferred_value = BLANK
            return
        self.set_timer(0.1, self.set_options)



"""Widget for setting application secrets and environment variables."""

from __future__ import annotations

import os

from rich.text import Text
from textual import on
from textual.app import ComposeResult
from textual.binding import Binding
from textual.containers import Horizontal, Vertical
from textual.events import Show
from textual.widgets import Button, Checkbox, Input, Label, Select, Static

from parllama.dialogs.yes_no_dialog import YesNoDialog
from parllama.secrets_manager import secrets_manager
from parllama.widgets.input_blur_submit import InputBlurSubmit


class SecretsView(Vertical):
    """Widget for setting application secrets and environment variables."""

    BINDINGS = [
        Binding(key="ctrl+c", action="app.copy_to_clipboard", show=True),
    ]

    DEFAULT_CSS = """
    SecretsView {
        width: 1fr;
        height: 1fr;
        overflow: auto;

        Horizontal {
            height: auto;
            Label {
                padding-top: 1;
                height: 3;
            }
        }

        .column {
            width: 1fr;
            height: auto;
        }

        .section {
            background: $panel;
            height: auto;
            width: 1fr;
            border: solid $primary;
            border-title-color: $primary;
        }
        InputBlurSubmit {
            width: 1fr;
        }
    }
    """

    def __init__(self, **kwargs) -> None:
        """Initialise the screen."""
        super().__init__(**kwargs)
        self.password_input = Input(value="", id="password", password=True)
        self.new_password_input = Input(value="", id="new_password", password=True)

    def compose(self) -> ComposeResult:  # pylint: disable=too-many-statements
        """Compose the content of the view."""
        with self.prevent(Input.Changed, Input.Submitted, Select.Changed, Checkbox.Changed):
            with Horizontal(classes="section") as vs:
                vs.border_title = Text.assemble(
                    "Vault: ",
                    (
                        "Locked" if secrets_manager.locked else "Unlocked",
                        "bold red" if secrets_manager.locked else "bold green",
                    ),
                )
                if not secrets_manager.has_password:
                    with Vertical(classes="height-auto plr-1"):
                        yield Label("Set Password")
                        yield self.password_input
                        yield Label("Verify Password")
                        yield self.new_password_input
                        yield Button("Set Password", id="set_password")
                else:
                    with Vertical(classes="height-auto plr-1"):
                        yield Label("Password")
                        yield self.password_input
                        yield Label("New Password")
                        yield self.new_password_input
                with Vertical(classes="height-auto plr-1"):
                    with Button(
                        "Import from ENV",
                        id="import_env",
                        disabled=secrets_manager.locked,
                    ) as b:
                        b.tooltip = "Import all matching environment variables from your environment."
                    yield Button("Clear Vault", id="clear_vault")
                    yield Static("Enter password and press enter to set password / unlock.")
                    yield Static("Blank password locks vault.")
                    yield Static("Enter password and new password to change password.")
                    yield Static("Mouse over fields to see corresponding environment variables.")

            with Horizontal():
                with Vertical(classes="column"):
                    with Vertical(classes="section") as vsf:
                        vsf.border_title = "API Keys"
                        with Horizontal():
                            yield Label("OpenAI")
                            yield InputBlurSubmit(
                                value=secrets_manager.get_secret("OPENAI_API_KEY", "", True),
                                id="OPENAI_API_KEY",
                                disabled=secrets_manager.locked,
                                tooltip="OPENAI_API_KEY",
                                classes="env-var",
                            )
                        with Horizontal():
                            yield Label("Groq")
                            yield InputBlurSubmit(
                                value=secrets_manager.get_secret("GROQ_API_KEY", "", True),
                                id="GROQ_API_KEY",
                                disabled=secrets_manager.locked,
                                tooltip="GROQ_API_KEY",
                                classes="env-var",
                            )
                        with Horizontal():
                            yield Label("Anthropic")
                            yield InputBlurSubmit(
                                value=secrets_manager.get_secret("ANTHROPIC_API_KEY", "", True),
                                id="ANTHROPIC_API_KEY",
                                disabled=secrets_manager.locked,
                                tooltip="ANTHROPIC_API_KEY",
                                classes="env-var",
                            )
                        with Horizontal():
                            yield Label("GoogleAI")
                            yield InputBlurSubmit(
                                value=secrets_manager.get_secret("GOOGLE_API_KEY", "", True),
                                id="GOOGLE_API_KEY",
                                disabled=secrets_manager.locked,
                                tooltip="GOOGLE_API_KEY",
                                classes="env-var",
                            )
                        with Horizontal():
                            yield Label("LangFlow")
                            yield InputBlurSubmit(
                                value=secrets_manager.get_secret("LANGFLOW_API_KEY", "", True),
                                id="LANGFLOW_API_KEY",
                                disabled=secrets_manager.locked,
                                tooltip="LANGFLOW_API_KEY",
                                classes="env-var",
                            )
                        with Horizontal():
                            yield Label("HuggingFace")
                            yield InputBlurSubmit(
                                value=secrets_manager.get_secret("HF_TOKEN", "", True),
                                id="HF_TOKEN",
                                disabled=secrets_manager.locked,
                                tooltip="HF_TOKEN",
                                classes="env-var",
                            )

                with Vertical(classes="section") as vs:
                    vs.border_title = "Search"
                    with Horizontal():
                        yield Label("Tavily")
                        yield InputBlurSubmit(
                            value=secrets_manager.get_secret("TAVILY_API_KEY", "", True),
                            id="TAVILY_API_KEY",
                            disabled=secrets_manager.locked,
                            tooltip="TAVILY_API_KEY",
                            classes="env-var",
                        )
                    with Horizontal():
                        yield Label("Serper")
                        yield InputBlurSubmit(
                            value=secrets_manager.get_secret("SERPER_API_KEY", "", True),
                            id="SERPER_API_KEY",
                            disabled=secrets_manager.locked,
                            tooltip="SERPER_API_KEY",
                            classes="env-var",
                        )
                    with Horizontal():
                        yield Label("Google Search")
                        yield InputBlurSubmit(
                            value=secrets_manager.get_secret("GOOGLE_CSE_ID", "", True),
                            id="GOOGLE_CSE_ID",
                            disabled=secrets_manager.locked,
                            tooltip="GOOGLE_CSE_ID",
                            classes="env-var",
                        )

    def _on_show(self, event: Show) -> None:
        """Handle show event"""
        self.screen.sub_title = (  # pylint: disable=attribute-defined-outside-init
            "Secrets"
        )
        self.refresh(recompose=True)

    @on(Select.Changed)
    def on_select_changed(self, event: Select.Changed) -> None:
        """Handle theme select changed"""
        event.stop()
        # ctrl: Select = event.control

    @on(Checkbox.Changed)
    def on_checkbox_changed(self, event: Checkbox.Changed) -> None:
        """Handle checkbox changed"""
        event.stop()
        # ctrl: Checkbox = event.control

    @on(Button.Pressed, "#import_env")
    async def import_env(self, event: Button.Pressed) -> None:
        """Import env vars into secrets"""
        event.stop()
        for env_input in self.query(".env-var"):
            if not isinstance(env_input, InputBlurSubmit):
                continue
            key: str = env_input.id or ""
            v = os.environ.get(key)
            if v:
                env_input.value = v
                await env_input.action_submit()
                self.notify(f"Imported: {key}")

    @on(Button.Pressed, "#clear_vault")
    def clear_vault(self, event: Button.Pressed) -> None:
        """Prompt to purge vault"""
        event.stop()
        self.app.push_screen(
            YesNoDialog(
                "Confirm Vault Purge",
                "All secrets will be cleared. Are you sure?",
                yes_first=False,
            ),
            self.confirm_clear_response,  # type: ignore
        )

    async def confirm_clear_response(self, res: bool) -> None:
        """Purge vault"""
        if not res:
            return
        secrets_manager.clear()
        await self.recompose()

    @on(Button.Pressed, "#set_password")
    async def set_password(self, event: Button.Pressed) -> None:
        """Set vault password"""
        event.stop()
        p1: str = self.password_input.value.strip()
        p2: str = self.new_password_input.value.strip()
        if not p1 or not p2:
            self.notify("Passwords cannot be blank", severity="error", timeout=8)
            return
        if p1 != p2:
            self.notify("Passwords do not match", severity="error", timeout=8)
            return
        secrets_manager.unlock(p1)
        self.notify("Password set")
        with self.prevent(Input.Changed):
            self.password_input.value = ""
            self.new_password_input.value = ""

        await self.recompose()

    # pylint: disable=too-many-branches,too-many-return-statements
    @on(Input.Submitted)
    async def on_input_submitted(self, event: Input.Submitted) -> None:
        """Handle input submission"""
        event.stop()
        if not secrets_manager.has_password:
            return
        ctrl: Input = event.control
        if event.validation_result is not None and not event.validation_result.is_valid:
            errors = ",".join([f.description or "Bad Value" for f in event.validation_result.failures])
            self.notify(f"{ctrl.id} [{errors}]", severity="error", timeout=8)
            return
        if ctrl.id == "password":
            try:
                v: str = ctrl.value.strip()
                if not v:
                    secrets_manager.lock()
                else:
                    secrets_manager.unlock(v)
                if secrets_manager.locked:
                    self.notify("Vault locked")
                else:
                    self.notify("Vault unlocked")
            except ValueError as e:
                self.notify(str(e), severity="error", timeout=8)
            finally:
                with self.prevent(Input.Changed):
                    self.password_input.value = ""
                await self.recompose()
            return
        if ctrl.id == "new_password":
            if not self.password_input.value or not self.new_password_input.value:
                return
            if self.password_input.value == self.new_password_input.value:
                self.notify("New password same as old", severity="error", timeout=8)
                return
            try:
                secrets_manager.change_password(self.password_input.value, self.new_password_input.value)
                self.notify("Password Changed")
            except ValueError as e:
                self.notify(str(e), severity="error", timeout=8)
            finally:
                with self.prevent(Input.Changed):
                    self.password_input.value = ""
                    self.new_password_input.value = ""
                await self.recompose()
            return
        if secrets_manager.locked:
            return
        secrets_manager[ctrl.id or ""] = ctrl.value



"""Local Model View"""

from __future__ import annotations

from functools import partial
from typing import cast

from textual import on
from textual.app import ComposeResult
from textual.binding import Binding
from textual.containers import Container, VerticalScroll
from textual.events import Focus, Show
from textual.screen import ScreenResultCallbackType
from textual.widget import Widget
from textual.widgets import Input, TabbedContent

from parllama.dialogs.input_dialog import InputDialog
from parllama.dialogs.model_details_dialog import ModelDetailsDialog
from parllama.dialogs.yes_no_dialog import YesNoDialog
from parllama.messages.messages import (
    LocalModelCopied,
    LocalModelCopyRequested,
    LocalModelDelete,
    LocalModelDeleted,
    LocalModelDeleteRequested,
    LocalModelListLoaded,
    LocalModelListRefreshRequested,
    LocalModelPulled,
    LocalModelPullRequested,
    LocalModelPushRequested,
    ModelInteractRequested,
    RegisterForUpdates,
    SetModelNameLoading,
    ShowLocalModel,
)
from parllama.ollama_data_manager import ollama_dm
from parllama.settings_manager import settings
from parllama.widgets.filter_input import FilterInput
from parllama.widgets.local_model_grid_list import LocalModelGridList
from parllama.widgets.local_model_list_item import LocalModelListItem


class LocalModelView(Container):
    """Local Model View"""

    BINDINGS = [
        Binding(
            key="ctrl+f",
            action="search_input_focus",
            description="Filter",
            show=True,
            priority=True,
        ),
        Binding(
            key="ctrl+r",
            action="refresh_models",
            description="Refresh",
            show=True,
            priority=True,
        ),
        Binding(
            key="p",
            action="pull_model",
            description="Pull",
            show=True,
        ),
        Binding(
            key="ctrl+p",
            action="pull_all_models",
            description="Pull All",
            show=True,
        ),
        Binding(
            key="ctrl+u",
            action="push_model",
            description="Push",
            show=True,
        ),
        Binding(
            key="ctrl+d",
            action="copy_model",
            description="Copy",
            show=True,
        ),
        Binding(
            key="ctrl+c",
            action="chat_model",
            description="Chat",
            show=True,
        ),
    ]
    DEFAULT_CSS = """
    LocalModelView {
        height: 1fr;
        width: 1fr;
        #search {
            margin-bottom: 1;
        }
        GridList {
            min-height: 1fr;
        }
    }
    """
    search_input: FilterInput

    def __init__(self, **kwargs) -> None:
        """Initialise the view"""
        super().__init__(**kwargs)
        self.sub_title = "Local Models"
        self.search_input = FilterInput(id="search", placeholder="Filter local models")
        self.grid = LocalModelGridList(id="grid_view")

    def _on_show(self, event: Show) -> None:
        """Focus the search on show"""
        self.screen.sub_title = "Local Models"

        with self.screen.prevent(TabbedContent.TabActivated):
            self.search_input.focus()

    def compose(self) -> ComposeResult:
        """Compose the Main screen."""
        with self.prevent(Focus, TabbedContent.TabActivated):
            yield self.search_input
            with VerticalScroll():
                with self.grid:
                    yield from ollama_dm.models

    async def on_mount(self) -> None:
        """Mount the view."""

        self.app.post_message(
            RegisterForUpdates(
                widget=self,
                event_names=[
                    "LocalModelListLoaded",
                    "LocalModelDeleted",
                    "SetModelNameLoading",
                    "LocalModelPulled",
                    "LocalModelPushed",
                ],
            )
        )
        if settings.load_local_models_on_startup:
            self.action_refresh_models()

    def action_refresh_models(self):
        """Refresh the models."""
        self.grid.loading = True
        self.post_message(LocalModelListRefreshRequested(widget=self))

    def action_pull_model(self) -> None:
        """Pull model"""
        if not self.grid.selected:
            return
        model_name: str = self.grid.selected.model.name
        self.grid.selected.loading = True
        self.post_message(LocalModelPullRequested(widget=self, model_name=model_name))

    def action_pull_all_models(self) -> None:
        """Pull all local models"""
        self.notify("Queuing pull of all local models...")
        for item in self.grid.query(LocalModelListItem):
            item.loading = True
            self.post_message(LocalModelPullRequested(widget=self, model_name=item.model.name, notify=False))

    def action_push_model(self) -> None:
        """Pull model"""
        if not self.grid.selected:
            return
        model_name: str = self.grid.selected.model.name
        self.grid.selected.loading = True
        self.post_message(LocalModelPushRequested(widget=self, model_name=model_name))

    @on(LocalModelListLoaded)
    def on_model_data_loaded(self, event: LocalModelListLoaded) -> None:
        """Rebuild model grid."""

        event.stop()

        # model_name: str = ""
        # if self.grid.selected:
        #     model_name = self.grid.selected.model.name

        to_remove: list[Widget] = []
        for child in self.grid.children:
            if isinstance(child, LocalModelListItem):
                to_remove.append(child)
        for child in to_remove:
            child.remove()

        self.grid.mount(*ollama_dm.models)
        self.grid.loading = False
        if self.search_input.value:
            self.grid.filter(self.search_input.value)
        # if self.parent and cast(Widget, self.parent).has_focus:
        #     if model_name:
        #         self.grid.select_by_name(model_name)
        #     else:
        #         self.grid.select_first_item()

    @on(LocalModelDeleteRequested)
    def on_model_delete_requested(self, event: LocalModelDeleteRequested) -> None:
        """Delete model requested."""
        event.stop()
        self.app.push_screen(
            YesNoDialog("Delete Model", "Delete model from local filesystem?", yes_first=False),
            cast(
                ScreenResultCallbackType[bool],
                partial(self.confirm_delete_response, event.model_name),
            ),
        )
        self.grid.set_item_loading(event.model_name, True)

    def confirm_delete_response(self, model_name: str, res: bool) -> None:
        """Confirm the deletion of a model."""
        if not res:
            self.grid.set_item_loading(model_name, False)
            return
        self.post_message(LocalModelDelete(model_name=model_name))

    @on(LocalModelDeleted)
    def on_model_deleted(self, event: LocalModelDeleted) -> None:
        """Model deleted remove item from grid."""
        event.stop()
        self.grid.remove_item(event.model_name)
        self.grid.action_select_left()

    @on(ShowLocalModel)
    def on_show_model(self, event: ShowLocalModel) -> None:
        """Show model details"""
        event.stop()
        self.app.push_screen(ModelDetailsDialog(event.model))

    @on(LocalModelPulled)
    def on_model_pulled(self, event: LocalModelPulled) -> None:
        """Model pulled turn off loading indicator."""
        event.stop()
        self.grid.set_item_loading(event.model_name, False)

    @on(Input.Changed, "#search")
    def on_search_input_changed(self, event: Input.Changed) -> None:
        """Handle search input change"""
        event.stop()
        self.grid.filter(event.value)

    def action_search_input_focus(self) -> None:
        """Focus the search input."""
        self.search_input.focus()

    def action_copy_model(self) -> None:
        """Copy local model"""
        if not self.grid.selected:
            return

        src_model_name: str = self.grid.selected.model.name
        self.app.push_screen(
            InputDialog(
                prompt="New model name",
                initial=src_model_name.split(":")[0],
                title="Copy Model",
                msg=f"Src: {src_model_name}",
            ),
            cast(
                ScreenResultCallbackType[str],
                partial(self.confirm_copy_response, src_model_name),
            ),
        )

    def confirm_copy_response(self, src_model_name: str, dst_model_name: str) -> None:
        """Get new name and confirm the copy of a model."""
        dst_model_name = dst_model_name.strip()
        if not dst_model_name:
            return
        self.post_message(
            LocalModelCopyRequested(
                widget=self,
                src_model_name=src_model_name,
                dst_model_name=dst_model_name,
            )
        )
        self.notify(f"Copy {src_model_name} to {dst_model_name}")

    @on(LocalModelCopied)
    def on_local_model_copied(self, event: LocalModelCopied) -> None:
        """Model copied event"""
        event.stop()

    @on(SetModelNameLoading)
    def on_set_model_name_loading(self, event: SetModelNameLoading) -> None:
        """Set model name loading"""
        event.stop()
        self.grid.set_item_loading(event.model_name, event.loading)

    def action_chat_model(self) -> None:
        """Chat with model"""
        if not self.grid.selected:
            return
        self.app.post_message(ModelInteractRequested(model_name=self.grid.selected.model.name))



"""Widget for managing custom prompts."""

from __future__ import annotations

from functools import partial
from typing import cast

from textual import on
from textual.app import ComposeResult
from textual.containers import Container, Horizontal, Vertical
from textual.events import Show
from textual.message import Message
from textual.screen import ScreenResultCallbackType
from textual.widgets import Button, Input, Label, Select

from parllama.chat_manager import chat_manager
from parllama.chat_prompt import ChatPrompt
from parllama.dialogs.edit_prompt_dialog import EditPromptDialog
from parllama.dialogs.import_fabric_dialog import ImportFabricDialog
from parllama.dialogs.yes_no_dialog import YesNoDialog
from parllama.messages.messages import DeletePrompt, PromptDeleteRequested, PromptSelected, RegisterForUpdates
from parllama.settings_manager import settings
from parllama.widgets.input_blur_submit import InputBlurSubmit
from parllama.widgets.prompt_list import PromptList
from parllama.widgets.provider_model_select import ProviderModelSelect


class PromptView(Container):
    """Widget for managing custom prompts."""

    DEFAULT_CSS = """
    PromptView {
        #tool_bar {
            height: 3;
            background: $surface-darken-1;
            Label {
                margin-top: 1;
                background: transparent;
            }
            #temperature_input {
                width: 12;
            }
        }
        #prompt_list {
            width: 1fr;
            height: 1fr;
            ListItem {
                height: 6;
            }
        }
    }
    """

    def __init__(self, **kwargs) -> None:
        """Initialise the screen."""
        super().__init__(**kwargs)
        self.list_view = PromptList(id="prompt_list")
        self.provider_model_select: ProviderModelSelect = ProviderModelSelect(
            classes="horizontal",
        )
        self.temperature_input: InputBlurSubmit = InputBlurSubmit(
            id="temperature_input",
            value=f"{settings.last_llm_config.temperature:.2f}",
            max_length=4,
            restrict=r"^\d?\.?\d?\d?$",
            valid_empty=False,
        )

    def compose(self) -> ComposeResult:
        """Compose the content of the view."""
        with Vertical():
            with Horizontal(id="tool_bar"):
                yield Button("New", id="new_prompt", variant="primary")
                yield self.provider_model_select
                yield Label("Temp")
                yield self.temperature_input
                yield Button("Import from Fabric", id="import")

            yield self.list_view

    def _on_show(self, event: Show) -> None:
        """Handle show event"""
        self.screen.sub_title = (  # pylint: disable=attribute-defined-outside-init
            "Prompts"
        )

    async def on_mount(self) -> None:
        """Set up the dialog once the DOM is ready."""
        self.app.post_message(
            RegisterForUpdates(
                widget=self,
                event_names=[
                    "DeletePrompt",
                    "PromptDeleteRequested",
                ],
            )
        )

    @on(DeletePrompt)
    def on_delete_prompt(self, event: DeletePrompt) -> None:
        """Handle deletion of a prompt."""
        event.stop()
        chat_manager.delete_prompt(event.prompt_id)

    @on(PromptDeleteRequested)
    def on_prompt_delete_requested(self, event: PromptDeleteRequested) -> None:
        """Delete model requested."""
        event.stop()
        self.app.push_screen(
            YesNoDialog("Confirm Delete", "Delete custom prompt?", yes_first=False),
            cast(
                ScreenResultCallbackType[bool],
                partial(self.confirm_delete_response, event.prompt_id),
            ),
        )

    def confirm_delete_response(self, prompt_id: str, res: bool) -> None:
        """Confirm the deletion of a model."""
        if not res:
            return
        self.post_message(DeletePrompt(prompt_id=prompt_id))

    @on(Button.Pressed, "#new_prompt")
    def action_new_prompt(self) -> None:
        """Open the new prompt screen."""
        prompt: ChatPrompt = ChatPrompt(name="New Prompt", description="")
        self.app.push_screen(
            EditPromptDialog(prompt),
            cast(ScreenResultCallbackType[bool], partial(self.do_add_prompt, prompt)),
        )

    def do_add_prompt(self, prompt: ChatPrompt, res: bool) -> None:
        """Add a new prompt to the list."""
        if not res:
            return

        chat_manager.add_prompt(prompt)
        self.notify("Prompt added")

    @on(Input.Submitted, "#temperature_input")
    def temperature_input_changed(self, event: Message) -> None:
        """Handle temperature input change"""
        event.stop()
        try:
            if self.temperature_input.value:
                chat_manager.prompt_temperature = float(self.temperature_input.value)
        except ValueError:
            pass

    @on(Select.Changed, "#model_name")
    def model_name_changed(self, event: Select.Changed) -> None:
        """Model name changed"""
        if event.value == Select.BLANK:
            chat_manager.prompt_llm_name = None
        else:
            chat_manager.prompt_llm_name = event.value  # type: ignore

    @on(PromptSelected)
    def on_prompt_selected(self, event: PromptSelected) -> None:
        """Prompt selected event"""
        try:
            if self.temperature_input.value:
                event.temperature = float(self.temperature_input.value)
        except ValueError:
            event.temperature = None
        if self.provider_model_select.is_valid():
            event.llm_provider = self.provider_model_select.provider_select.value  # type: ignore
            event.model_name = self.provider_model_select.model_select.value  # type: ignore
        else:
            event.model_name = None

    @on(Button.Pressed, "#import")
    def import_prompts(self, event: Message) -> None:
        """Handle import prompts"""
        event.stop()
        self.app.push_screen(ImportFabricDialog())



"""Manager for application secrets."""

from __future__ import annotations

import base64
import os
from pathlib import Path
from typing import Any

import orjson as json
from cryptography.exceptions import InvalidTag
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC
from orjson import JSONDecodeError
from textual.app import App

from parllama.par_event_system import ParEventSystemBase
from parllama.settings_manager import settings


class SecretsManager(ParEventSystemBase):
    """
    Manager for application secrets.
    Uses PBKDF2 with HMAC-SHA256 to derive a key from a password,
    then uses AES-GCM encryption for storing secrets.
    Secrets are never stored in plain text, and only decrypted when accessed.
    """

    _key_secure: str | None = None
    """Encrypted current vault password"""
    _key: bytes | None = None
    """Decryption key derived from the password."""
    _salt: bytes
    """Used with password to derive the en/decryption key."""
    _secrets: dict[str, str]
    """Dict containing encrypted secrets."""
    _secrets_file: Path
    """JSON file containing encrypted secrets."""

    def __init__(self, secrets_file: str, **kwargs) -> None:
        """Initialize Manager and load vault."""
        super().__init__(**kwargs)
        self._secrets = {}
        self._salt = gen_salt()
        self._key_secure = None
        self._key = None
        self._secrets_file = Path(secrets_file)

    def set_app(self, app: App[Any] | None) -> None:
        """Set the app and load existing sessions and prompts from storage"""
        super().set_app(app)
        self._load_secrets()
        env_key = os.environ.get("PARLLAMA_VAULT_KEY")
        if env_key:
            self.unlock(env_key, True)

    def _load_secrets(self) -> None:
        """Load secrets from the secrets file."""
        try:
            data = json.loads(self._secrets_file.read_bytes())
            self._salt = base64.b64decode(data.get("__salt__"))
            self._key_secure = data.get("__key__")
            self._secrets = data.get("secrets", {})
        except FileNotFoundError:
            self._secrets = {}
            self._salt = gen_salt()
            self._key_secure = None
        except JSONDecodeError as e:
            self.log_it(e)
            self.log_it("Invalid secrets file format", notify=True, severity="error")
            raise ValueError("Invalid secrets file format") from e

    def _save_secrets(self) -> None:
        """Saves secrets to the secrets file."""
        data = {
            "__salt__": base64.b64encode(self._salt).decode("utf-8"),
            "__key__": self._key_secure,
            "secrets": self._secrets,
        }
        self._secrets_file.write_bytes(json.dumps(data, str, json.OPT_INDENT_2))
        self.import_to_env(True)

    def _derive_key(self, password: str, alt_salt: bytes | None = None) -> bytes:
        """Derives a key from the given password and the stored salt."""
        return derive_key(password, alt_salt or self._salt)

    @property
    def locked(self) -> bool:
        """Checks if the secrets manager is locked."""
        return self._key is None

    def lock(self) -> None:
        """Locks the secrets manager."""
        self._key = None

    @property
    def has_password(self) -> bool:
        """Checks if a password is set."""
        return self._key_secure is not None

    def change_password(self, old_password: str, new_password: str, no_raise: bool = False) -> None:
        """Changes the password and re-encrypts existing secrets."""
        if not self.has_password:
            self.unlock(new_password)
            return

        if not self.verify_password(old_password):
            if no_raise:
                self.log_it("Invalid old password.", notify=True, severity="error")
                return
            raise ValueError("Invalid old password.")
        if new_password == old_password:
            return
        self.unlock(old_password)
        new_key: bytes = self._derive_key(new_password)
        self._key_secure = self._encrypt(new_password, new_key)

        encrypted_secrets = {}
        for key, value in self._secrets.items():
            encrypted_secrets[key] = self._encrypt(self._decrypt(value), new_key)
        self._secrets = encrypted_secrets
        self._key = new_key
        self._save_secrets()

    def verify_password(self, password: str) -> bool:
        """Verifies the given password. Returns True if vault password is not set."""
        try:
            if not self.has_password:
                return True
            return (
                self._decrypt(self._key_secure, self._derive_key(password)) == password  # type: ignore
            )
        except ValueError as e:
            self.log_it(e)
            return False

    def unlock(self, password: str, no_raise: bool = False) -> bool:
        """Unlock the vault or set initial vault password."""
        if not self.verify_password(password):
            self.lock()
            if no_raise:
                self.log_it("Invalid password.", notify=True, severity="error")
                return False
            raise ValueError("Invalid password.")
        self._key = self._derive_key(password)
        if self._key_secure is None:
            self._key_secure = self._encrypt(password, self._derive_key(password))
            self._save_secrets()
        return True

    def _encrypt(self, plaintext: str, alt_key: bytes | None = None) -> str:
        key: bytes | None = alt_key or self._key
        if key is None:
            raise ValueError("Password not set. Use unlock() before encrypting.")
        return encrypt(plaintext, key)

    def _decrypt(self, ciphertext: str, alt_key: bytes | None = None) -> str:
        """decrypt ciphertext with the provided key"""
        if self._key is None and alt_key is None:
            raise ValueError("Vault locked. Use unlock() before decrypting.")
        return decrypt(ciphertext, alt_key or self._key)  # type: ignore

    def add_secret(self, key: str, value: str, no_raise: bool = False):
        """Adds a new secret, encrypts it, and saves it to the file."""
        if not self._key:
            if no_raise:
                self.log_it("Password not set. Use unlock() before adding a secret.")
                return
            raise ValueError("Password not set. Use unlock() before adding a secret.")

        encrypted_value = self._encrypt(value)
        self._secrets[key] = encrypted_value
        self._save_secrets()

    def get_secret(self, key: str, default: str | None = None, no_raise: bool = False) -> str:
        """Decrypts and returns the secret associated with the given key."""
        if self.locked:
            if no_raise:
                return "Vault is locked"
            raise ValueError("Vault is locked")

        encrypted_value = self._secrets.get(key)
        if encrypted_value is None:
            if default is None:
                if no_raise:
                    return ""
                raise KeyError(f"No secret found for key: {key}")
            return default
        try:
            return self._decrypt(encrypted_value)
        except ValueError as e:
            if no_raise:
                return default or ""
            raise e

    def get_secret_with_pw(self, key: str, password: str, no_raise: bool = False) -> str:
        """Returns secret associated with the given key, using the provided password if vault is locked."""
        try:
            if self.locked:
                if not password:
                    if no_raise:
                        self.log_it(
                            "Password required to unlock the vault.",
                            notify=True,
                            severity="error",
                        )
                        return ""
                    raise ValueError("Password required to unlock the vault.")
                self.unlock(password)
            return self.get_secret(key)
        except ValueError as e:
            if no_raise:
                return ""
            raise e

    def encrypt_with_password(self, plaintext: str, password: str) -> str:
        """Encrypts plaintext with the provided password."""
        return self._encrypt(plaintext, self._derive_key(password))

    def decrypt_with_password(self, ciphertext: str, password: str) -> str:
        """Decrypts ciphertext with the provided password."""
        return self._decrypt(ciphertext, self._derive_key(password))

    def remove_secret(self, key: str, no_raise: bool = False) -> None:
        """Removes the secret associated with the given key and saves the changes."""
        if key in self._secrets:
            del self._secrets[key]
            self._save_secrets()
        else:
            if no_raise:
                self.log_it(f"No secret found for key: {key}", notify=True, severity="warning")
                return
            raise KeyError(f"No secret found for key: {key}")

    def clear(self) -> None:
        """Clear vault and remove password"""
        self._secrets.clear()
        self._salt = gen_salt()
        self._key_secure = None
        self.lock()
        self._save_secrets()
        self.log_it("Vault cleared and password removed.", notify=True)

    def import_to_env(self, no_raise: bool = False) -> None:
        """Imports secrets from the secrets file to the environment variables."""
        if self.locked:
            if no_raise:
                return
            raise ValueError("Vault is locked")
        for key, value in self._secrets.items():
            v = self._decrypt(value)
            if v:
                os.environ[key] = v

    @property
    def is_empty(self) -> bool:
        """Checks if the secrets manager is empty."""
        return len(self._secrets) == 0

    def __len__(self):
        """Returns the number of secrets stored."""
        return len(self._secrets)

    def __getitem__(self, key: str) -> str:
        """Allows access to secrets using square bracket notation."""
        return self.get_secret(key)

    def __setitem__(self, key: str, value: str) -> None:
        """Allows setting a secret using square bracket notation."""
        self.add_secret(key, value)

    def __delitem__(self, key: str) -> None:
        """Allows deletion of a secret using square bracket notation."""
        self.remove_secret(key)

    def __contains__(self, key: str) -> bool:
        """Allows checking if a key exists using the `in` keyword."""
        return key in self._secrets


def derive_key(password: str, salt: bytes) -> bytes:
    """Derives a key from the given password and salt."""
    kdf = PBKDF2HMAC(
        algorithm=hashes.SHA256(),
        length=32,
        salt=salt,
        iterations=100000,
        backend=default_backend(),
    )
    return kdf.derive(password.encode("utf-8"))


def encrypt(plaintext: str, key: bytes) -> str:
    """Encrypts plaintext with the given key bytes."""
    try:
        iv = os.urandom(12)
        cipher = Cipher(algorithms.AES(key), modes.GCM(iv), backend=default_backend())
        encryptor = cipher.encryptor()
        encrypted = encryptor.update(plaintext.encode("utf-8")) + encryptor.finalize()
        return base64.b64encode(iv + encryptor.tag + encrypted).decode("utf-8")
    except (TypeError, ValueError, AttributeError) as e:
        raise ValueError(f"An error occurred: {e}") from e


def decrypt(ciphertext: str, key: bytes) -> str:
    """Decrypts ciphertext with the given key bytes."""
    try:
        decoded = base64.b64decode(ciphertext)
        iv, tag, encrypted = decoded[:12], decoded[12:28], decoded[28:]
        cipher = Cipher(
            algorithms.AES(key),
            modes.GCM(iv, tag),
            backend=default_backend(),
        )
        decryptor = cipher.decryptor()
        plaintext = decryptor.update(encrypted) + decryptor.finalize()
        return plaintext.decode("utf-8")
    except (ValueError, TypeError, InvalidTag) as e:
        raise ValueError("Bad key or invalid / corrupted ciphertext.") from e


def encrypt_with_password(plaintext: str, password: str, salt: str | bytes) -> str:
    """Encrypts plaintext with the provided password."""
    return encrypt(
        plaintext,
        derive_key(password, base64.b64decode(salt) if isinstance(salt, str) else salt),
    )


def decrypt_with_password(ciphertext: str, password: str, salt: str | bytes) -> str:
    """Decrypts ciphertext with the provided password and salt."""
    return decrypt(
        ciphertext,
        derive_key(password, base64.b64decode(salt) if isinstance(salt, str) else salt),
    )


def gen_salt() -> bytes:
    """Generates random salt bytes."""
    return os.urandom(16)


secrets_manager = SecretsManager(settings.secrets_file)



"""Import Fabric prompts from fabric repo."""

from __future__ import annotations

import hashlib
import os
import shutil
import tempfile
import zipfile

import requests

from parllama.chat_manager import chat_manager
from parllama.chat_message import ParllamaChatMessage
from parllama.chat_prompt import ChatPrompt
from parllama.par_event_system import ParEventSystemBase
from parllama.settings_manager import settings


class ImportFabricManager(ParEventSystemBase):
    """Import Fabric prompts from fabric repo."""

    id_to_prompt: dict[str, ChatPrompt]
    prompts: list[ChatPrompt]
    import_ids: set[str]

    def __init__(self) -> None:
        """Initialize the import manager."""
        super().__init__(id="import_fabric_manager")
        self.prompts = []
        self.id_to_prompt = {}
        self.import_ids = set()
        self.repo_zip_url = "https://github.com/danielmiessler/fabric/archive/refs/heads/main.zip"
        self._cache_folder = os.path.join(settings.cache_dir, "fabric_prompts")
        self._last_folder: str | None = None

    def import_patterns(self) -> None:
        """Import requested Fabric prompts."""
        for prompt_id in self.import_ids:
            prompt = self.id_to_prompt.get(prompt_id)
            if not prompt:
                continue
            prompt.source = "fabric"
            chat_manager.add_prompt(prompt)
            prompt.is_dirty = True
            prompt.save()

    def fetch_patterns(self, force: bool = False) -> None:
        """Create prompts from GitHub zip file."""

        if os.path.exists(self._cache_folder):
            if force:
                shutil.rmtree(self._cache_folder)
            else:
                return

        with tempfile.TemporaryDirectory() as temp_dir:
            self._last_folder = temp_dir
            zip_path = os.path.join(temp_dir, "repo.zip")
            self.download_zip(self.repo_zip_url, zip_path)
            extracted_folder_path = self.extract_zip(zip_path, temp_dir)
            # The patterns folder will be inside "fabric-main" after extraction
            patterns_source_path = os.path.join(extracted_folder_path, "fabric-main", "patterns")
            if not os.path.exists(patterns_source_path):
                raise FileNotFoundError("Patterns folder not found in the downloaded zip.")
            shutil.copytree(patterns_source_path, self._cache_folder)

    def read_patterns(self, force: bool = False) -> list[ChatPrompt]:
        """Read prompts from cache."""

        self.prompts.clear()
        self.id_to_prompt.clear()
        self.import_ids.clear()

        try:
            if not os.path.exists(self._cache_folder) or force:
                self.fetch_patterns(force)
        except FileNotFoundError:
            return []

        if not os.path.exists(self._cache_folder):
            return []

        pattern_folder_list = os.listdir(self._cache_folder)
        for pattern_name in pattern_folder_list:
            src_prompt_path = os.path.join(self._cache_folder, pattern_name, "system.md")
            if not os.path.exists(src_prompt_path):
                continue
            with open(src_prompt_path, encoding="utf-8") as f:
                prompt_content = ""
                for line in f.readlines():
                    if line.upper().startswith("# INPUT") or line.upper().startswith("INPUT:"):
                        break
                    prompt_content += line + "\n"
                prompt_content = prompt_content.strip()
                prompt: ChatPrompt = self.markdown_to_prompt(pattern_name, prompt_content)
                self.prompts.append(prompt)
                self.id_to_prompt[prompt.id] = prompt
        return self.prompts

    def markdown_to_prompt(self, pattern_name: str, prompt_content: str) -> ChatPrompt:
        """Convert markdown to ChatPrompt."""
        description = self.get_description(prompt_content)
        prompt = ChatPrompt(
            id=hashlib.md5(prompt_content.encode()).hexdigest(),
            name=pattern_name,
            description=description,
            messages=[ParllamaChatMessage(role="system", content=prompt_content)],
            source="fabric",
        )
        return prompt

    @staticmethod
    def get_description(prompt_data: str) -> str:
        """Extract description from prompt data."""
        started = False
        for line in prompt_data.split("\n"):
            line = line.strip()
            if line.startswith("# IDENTITY and PURPOSE"):
                started = True
                continue
            if not started or not line:
                continue
            return line
        return ""

    @staticmethod
    def download_zip(url: str, save_path: str) -> None:
        """Download the zip file from the specified URL."""
        response = requests.get(url, timeout=10)
        response.raise_for_status()  # Check if the download was successful
        with open(save_path, "wb") as f:
            f.write(response.content)

    @staticmethod
    def extract_zip(zip_path: str, extract_to: str) -> str:
        """Extract the zip file to the specified directory."""
        with zipfile.ZipFile(zip_path, "r") as zip_ref:
            zip_ref.extractall(extract_to)
        print("Extracted zip file successfully.")
        return extract_to  # Return the path to the extracted contents

    def test_import(self) -> None:
        """Test importing fabric prompts."""
        with open(
            "d:/repos/parllama/fabric_samples/extract_wisdom/system.md",
            encoding="utf-8",
        ) as f:
            prompt_content = ""
            for line in f.readlines():
                if line.upper().startswith("# INPUT") or line.upper().startswith("INPUT:"):
                    break
                prompt_content += line + "\n"
            prompt_content = prompt_content.strip()
            prompt: ChatPrompt = self.markdown_to_prompt("extract_wisdom", prompt_content)
        chat_manager.add_prompt(prompt)
        prompt.is_dirty = True
        prompt.save()
        self.log_it("Prompt imported: extract_wisdom", notify=True)


import_fabric_manager = ImportFabricManager()
